#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List
import numpy as np

from .base import BaseReduceRouter, BaseTopkReduceRouter, BaseEmbedReduceRouter
from ..proto import blob2array


class DocFillReducer(BaseReduceRouter):
    """
    Gather all documents raw content from multiple shards.
    This is only useful when you have
    - multiple doc-indexer and docs are spreaded over multiple shards.
    - require full-doc retrieval with the original content, not just an doc id
    Ideally, only each doc can only belong to one shard.
    """
    def apply(self, msg: 'gnes_pb2.Message', accum_msgs: List['gnes_pb2.Message'], *args, **kwargs):
        final_docs = []
        for idx in range(len(msg.response.search.topk_results)):
            # get result from all shards, some may return None, we only take the first non-None doc
            final_docs.append([m.response.search.topk_results[idx] for m in accum_msgs if
                               m.response.search.topk_results[idx].doc.WhichOneof('raw_data') is not None][0])
        msg.response.search.ClearField('topk_results')
        msg.response.search.topk_results.extend(final_docs)

        super().apply(msg, accum_msgs)


class DocTopkReducer(BaseTopkReduceRouter):
    """
    Gather all docs by their doc_id, result in a topk doc list
    """

    def get_key(self, x: 'gnes_pb2.Response.QueryResponse.ScoredResult') -> str:
        return x.doc.doc_id

    def set_key(self, x: 'gnes_pb2.Response.QueryResponse.ScoredResult', k: str):
        x.doc.doc_id = k


class Chunk2DocTopkReducer(BaseTopkReduceRouter):
    """
    Gather all chunks by their doc_id, result in a topk doc list.
    This is almost always useful, as the final result should be group by doc_id
    not chunk
    """

    def get_key(self, x: 'gnes_pb2.Response.QueryResponse.ScoredResult') -> str:
        return x.chunk.doc_id

    def set_key(self, x: 'gnes_pb2.Response.QueryResponse.ScoredResult', k: str):
        x.doc.doc_id = k


class ChunkTopkReducer(BaseTopkReduceRouter):
    """
    Gather all chunks by their chunk_id from all shards, aka doc_id-offset, result in a topk chunk list
    """

    def get_key(self, x: 'gnes_pb2.Response.QueryResponse.ScoredResult') -> str:
        return '%d-%d' % (x.chunk.doc_id, x.chunk.offset)

    def set_key(self, x: 'gnes_pb2.Response.QueryResponse.ScoredResult', k: str):
        x.chunk.doc_id, x.chunk.offset = map(int, k.split('-'))


class ConcatEmbedRouter(BaseEmbedReduceRouter):
    """
    Gather all embeddings from multiple encoders and concat them on a specific axis.
    In default, concat will happen on the last axis.
    chunk_idx, doc_idx denote index in for loop used in BaseEmbedReduceRouter
    """

    def reduce_embedding(self, accum_msgs: List['gnes_pb2.Message'], msg_type: str, chunk_idx: int, doc_idx: int):
        if msg_type == 'query':
            return np.concatenate([blob2array(m.request.search.query.chunks[chunk_idx].embedding)
                                   for m in accum_msgs], axis=1)
        elif msg_type == 'index':
            return np.concatenate([blob2array(m.request.index.docs[doc_idx].chunks[chunk_idx].embedding)
                                   for m in accum_msgs], axis=1)
        else:
            self.logger.error('dont know how to handle %s' % msg_type)


class AvgEmbedRouter(BaseEmbedReduceRouter):
    """
    Gather all embeddings from multiple encoders and do average on a specific axis.
    In default, average will happen on the first axis.
    chunk_idx, doc_idx denote index in for loop used in BaseEmbedReduceRouter
    """

    def reduce_embedding(self, accum_msgs: List['gnes_pb2.Message'], msg_type: str, chunk_idx: int, doc_idx: int):
        if msg_type == 'query':
            return np.mean([blob2array(m.request.search.query.chunks[chunk_idx].embedding)
                                   for m in accum_msgs], axis=0)
        elif msg_type == 'index':
            return np.mean([blob2array(m.request.index.docs[doc_idx].chunks[chunk_idx].embedding)
                                   for m in accum_msgs], axis=0)
        else:
            self.logger.error('dont know how to handle %s' % msg_type)


import copy
import os
import time
import unittest

from gnes.cli.api import healthcheck
from gnes.cli.parser import set_router_parser, set_healthcheck_parser
from gnes.service.base import ServiceManager
from gnes.service.router import RouterService


class TestHealthCheck(unittest.TestCase):

    def test_health_check(self):
        a = set_router_parser().parse_args([
            '--yaml_path', 'BaseRouter',
        ])
        a1 = copy.deepcopy(a)
        b = set_healthcheck_parser().parse_args([
            '--port', str(a.port_ctrl)
        ])

        # before - fail
        with self.assertRaises(SystemExit) as cm:
            healthcheck(b)

        self.assertEqual(cm.exception.code, 1)

        # running - success
        with self.assertRaises(SystemExit) as cm:
            with RouterService(a):
                time.sleep(2)
                healthcheck(b)
        self.assertEqual(cm.exception.code, 0)

        # running - managerservice - success
        with self.assertRaises(SystemExit) as cm:
            with ServiceManager(RouterService, a1):
                time.sleep(2)
                healthcheck(b)
        self.assertEqual(cm.exception.code, 0)

        # after - fail
        with self.assertRaises(SystemExit) as cm:
            healthcheck(b)

        self.assertEqual(cm.exception.code, 1)

    @unittest.SkipTest
    def test_hc_os_env(self):
        os.environ['GNES_CONTROL_PORT'] = str(56789)
        a = set_router_parser().parse_args([
            '--yaml_path', 'BaseRouter',
        ])
        a1 = copy.deepcopy(a)
        b = set_healthcheck_parser().parse_args([
            '--port', os.environ.get('GNES_CONTROL_PORT')
        ])

        # before - fail
        with self.assertRaises(SystemExit) as cm:
            healthcheck(b)

        self.assertEqual(cm.exception.code, 1)

        # running - success
        with self.assertRaises(SystemExit) as cm:
            with RouterService(a):
                time.sleep(2)
                healthcheck(b)
        self.assertEqual(cm.exception.code, 0)

        # running - managerservice - success
        with self.assertRaises(SystemExit) as cm:
            with ServiceManager(RouterService, a1):
                time.sleep(2)
                healthcheck(b)
        self.assertEqual(cm.exception.code, 0)

        # after - fail
        with self.assertRaises(SystemExit) as cm:
            healthcheck(b)

        self.assertEqual(cm.exception.code, 1)
        os.unsetenv('GNES_CONTROL_PORT')



import os
import unittest

from gnes.encoder.base import BaseEncoder, PipelineEncoder
from gnes.helper import train_required


class DummyEncoder(BaseEncoder):

    def train(self, *args, **kwargs):
        self.logger.info('you just trained me!')
        pass

    @train_required
    def encode(self, x):
        return x + 1


class TestPipeTrain(unittest.TestCase):
    def setUp(self):
        self.dirname = os.path.dirname(__file__)

    def test_train(self):
        de = DummyEncoder()
        self.assertRaises(RuntimeError, de.encode, 1)
        de.train()
        self.assertEqual(2, de.encode(1))

    def tearDown(self):
        if os.path.exists('dummy-pipeline.bin'):
            os.remove('dummy-pipeline.bin')
        if os.path.exists('dummy-pipeline.yml'):
            os.remove('dummy-pipeline.yml')

    def test_pipeline_train(self):
        p = PipelineEncoder()
        p.components = lambda: [DummyEncoder(), DummyEncoder(), DummyEncoder()]
        self.assertRaises(RuntimeError, p.encode, 1)
        p.train(1)
        self.assertEqual(4, p.encode(1))
        p.name = 'dummy-pipeline'
        p.dump()
        p.dump_yaml()
        a = BaseEncoder.load_yaml(p.yaml_full_path)
        self.assertEqual(4, a.encode(1))

    @unittest.SkipTest
    def test_load_yaml(self):
        p = BaseEncoder.load_yaml(os.path.join(self.dirname, 'yaml', 'pipeline-multi-encoder.yml'))
        self.assertRaises(RuntimeError, p.encode, 1)
        p.train(1)
        self.assertEqual(5, p.encode(1))
        p = BaseEncoder.load_yaml(os.path.join(self.dirname, 'yaml', 'pipeline-multi-encoder.yml'))
        self.assertRaises(RuntimeError, p.encode, 1)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from collections import defaultdict
from functools import wraps
from typing import List, Any, Union, Callable, Tuple

import numpy as np

from ..base import TrainableBase, CompositionalTrainableBase
from ..proto import gnes_pb2, blob2array
from ..score_fn.base import get_unary_score, ModifierScoreFn


class BaseIndexer(TrainableBase):
    def __init__(self,
                 normalize_fn: 'BaseScoreFn' = None,
                 score_fn: 'BaseScoreFn' = None,
                 is_big_score_similar: bool = False,
                 *args, **kwargs):
        """
        Base indexer, a valid indexer must implement :py:meth:`add` and :py:meth:`query` methods

        :type score_fn: advanced score function
        :type normalize_fn: normalizing score function
        :type is_big_score_similar: when set to true, then larger score means more similar
        """
        super().__init__(*args, **kwargs)
        self.normalize_fn = normalize_fn if normalize_fn else ModifierScoreFn()
        self.score_fn = score_fn if score_fn else ModifierScoreFn()
        self.normalize_fn._context = self
        self.score_fn._context = self
        self.is_big_score_similar = is_big_score_similar
        self._num_docs = 0
        self._num_chunks = 0
        self._num_chunks_in_doc = defaultdict(int)

    def add(self, keys: Any, docs: Any, weights: List[float], *args, **kwargs):
        pass

    def query(self, keys: Any, *args, **kwargs) -> List[Any]:
        pass

    def query_and_score(self, q_chunks: List[Union['gnes_pb2.Chunk', 'gnes_pb2.Document']], top_k: int) -> List[
        'gnes_pb2.Response.QueryResponse.ScoredResult']:
        raise NotImplementedError

    @property
    def num_docs(self):
        return self._num_docs

    @property
    def num_chunks(self):
        return self._num_chunks


class BaseChunkIndexer(BaseIndexer):
    """Storing chunks and their vector representations """

    def __init__(self, helper_indexer: 'BaseChunkIndexerHelper' = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper_indexer = helper_indexer

    def add(self, keys: List[Tuple[int, int]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
        """
        adding new chunks and their vector representations

        :param keys: list of (doc_id, offset) tuple
        :param vectors: vector representations
        :param weights: weight of the chunks
        """
        pass

    def query(self, keys: np.ndarray, top_k: int, *args, **kwargs) -> List[List[Tuple]]:
        pass

    def query_and_score(self, q_chunks: List['gnes_pb2.Chunk'], top_k: int, *args, **kwargs) -> List[
        'gnes_pb2.Response.QueryResponse.ScoredResult']:
        vecs = [blob2array(c.embedding) for c in q_chunks]
        queried_results = self.query(np.stack(vecs), top_k=top_k)
        results = []
        for q_chunk, topk_chunks in zip(q_chunks, queried_results):
            for _doc_id, _offset, _weight, _relevance in topk_chunks:
                r = gnes_pb2.Response.QueryResponse.ScoredResult()
                r.chunk.doc_id = _doc_id
                r.chunk.offset = _offset
                r.chunk.weight = _weight
                _score = get_unary_score(value=_relevance,
                                         name=self.__class__.__name__,
                                         operands=[
                                             dict(name='doc_chunk',
                                                  doc_id=_doc_id,
                                                  offset=_offset),
                                             dict(name='query_chunk',
                                                  offset=q_chunk.offset)])
                _score = self.normalize_fn(_score)
                _score = self.score_fn(_score, q_chunk, r.chunk, queried_results)
                r.score.CopyFrom(_score)
                results.append(r)
        return results

    @staticmethod
    def update_counter(func):
        @wraps(func)
        def arg_wrapper(self, keys: List[Tuple[int, int]], *args, **kwargs):
            doc_ids, _ = zip(*keys)
            self._num_docs += len(set(doc_ids))
            self._num_chunks += len(keys)
            for doc_id in doc_ids:
                self._num_chunks_in_doc[doc_id] += 1
            return func(self, keys, *args, **kwargs)

        return arg_wrapper

    @staticmethod
    def update_helper_indexer(func):
        @wraps(func)
        def arg_wrapper(self, keys: List[Tuple[int, int]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
            r = func(self, keys, vectors, weights, *args, **kwargs)
            if self.helper_indexer:
                self.helper_indexer.add(keys, weights, *args, **kwargs)
            return r

        return arg_wrapper

    @property
    def num_docs(self):
        if self.helper_indexer:
            return self.helper_indexer._num_docs
        else:
            return self._num_docs

    @property
    def num_chunks(self):
        if self.helper_indexer:
            return self.helper_indexer._num_chunks
        else:
            return self._num_chunks

    def num_chunks_in_doc(self, doc_id: int):
        if self.helper_indexer:
            return self.helper_indexer._num_chunks_in_doc[doc_id]
        else:
            self.logger.warning('enable helper_indexer to track num_chunks_in_doc')


class BaseDocIndexer(BaseIndexer):
    """Storing documents and contents """

    def add(self, keys: List[int], docs: List['gnes_pb2.Document'], *args, **kwargs):
        """
        adding new docs and their protobuf representation

        :param keys: list of doc_id
        :param docs: list of protobuf Document objects
        """
        pass

    def query(self, keys: List[int], *args, **kwargs) -> List['gnes_pb2.Document']:
        pass

    def query_and_score(self, docs: List['gnes_pb2.Response.QueryResponse.ScoredResult'], *args, **kwargs) -> List[
        'gnes_pb2.Response.QueryResponse.ScoredResult']:
        keys = [r.doc.doc_id for r in docs]
        results = []
        queried_results = self.query(keys, *args, **kwargs)
        for d, r in zip(queried_results, docs):
            if d:
                r.doc.CopyFrom(d)
                _score = self.normalize_fn(r.score)
                _score = self.score_fn(_score, d)
                r.score.CopyFrom(_score)
            results.append(r)
        return results

    @staticmethod
    def update_counter(func):
        @wraps(func)
        def arg_wrapper(self, keys: List[int], docs: List['gnes_pb2.Document'], *args, **kwargs):
            self._num_docs += len(keys)
            self._num_chunks += sum(len(d.chunks) for d in docs)
            return func(self, keys, docs, *args, **kwargs)

        return arg_wrapper


class BaseChunkIndexerHelper(BaseChunkIndexer):
    """A helper class for storing chunk info, doc mapping, weights.
    This is especially useful when ChunkIndexer can not store these information by itself
    """

    def add(self, keys: List[Tuple[int, int]], weights: List[float], *args, **kwargs) -> int:
        pass

    def query(self, keys: List[int], *args, **kwargs) -> List[Tuple[int, int, float]]:
        pass


class JointIndexer(CompositionalTrainableBase):

    @property
    def components(self):
        return self._component

    @components.setter
    def components(self, comps: Callable[[], Union[list, dict]]):
        if not callable(comps):
            raise TypeError('component must be a callable function that returns '
                            'a List[BaseIndexer]')
        if not getattr(self, 'init_from_yaml', False):
            self._component = comps()
        else:
            self.logger.info('component is omitted from construction, '
                             'as it is initialized from yaml config')

        self._binary_indexer = None
        self._doc_indexer = None
        for c in self.components:
            if isinstance(c, BaseChunkIndexer):
                self._binary_indexer = c
            elif isinstance(c, BaseDocIndexer):
                self._doc_indexer = c
        if not self._binary_indexer or not self._doc_indexer:
            raise ValueError('"JointIndexer" requires a valid pair of "BaseBinaryIndexer" and "BaseTextIndexer"')

    def add(self, keys: Any, docs: Any, *args,
            **kwargs) -> None:
        if isinstance(docs, np.ndarray):
            self._binary_indexer.add(keys, docs, *args, **kwargs)
        elif isinstance(docs, list):
            self._doc_indexer.add(keys, docs, *args, **kwargs)
        else:
            raise TypeError('can not find an indexer for doc type: %s' % type(docs))

    def query(self,
              keys: Any,
              top_k: int,
              *args,
              **kwargs) -> List[List[Tuple]]:
        topk_results = self._binary_indexer.query(keys, top_k, *args, **kwargs)
        doc_caches = dict()
        topk_results_with_docs = []
        for topk in topk_results:
            topk_wd = []
            for doc_id, offset, weight, score in topk:
                doc = doc_caches.get(doc_id, self._doc_indexer.query([doc_id])[0])
                doc_caches[doc_id] = doc
                topk_wd.append((doc_id, offset, weight, score, doc.chunks[offset]))
            topk_results_with_docs.append(topk_wd)
        return topk_results_with_docs



import copy
from collections import OrderedDict, defaultdict
from contextlib import ExitStack
from typing import Union, Tuple, List, Optional, Iterator

from .helper import *
from ..base import TrainableBase
from ..helper import set_logger
from ..service.base import SocketType, BaseService


class Flow(TrainableBase):
    """
    GNES Flow: an intuitive way to build workflow for GNES.

    You can use :py:meth:`.add()` then :py:meth:`.build()` to customize your own workflow.
    For example:

    .. highlight:: python
    .. code-block:: python

        from gnes.flow import Flow, Service as gfs

        f = (Flow(check_version=False, route_table=True)
             .add(gfs.Preprocessor, yaml_path='BasePreprocessor')
             .add(gfs.Encoder, yaml_path='BaseEncoder')
             .add(gfs.Router, yaml_path='BaseRouter'))

        with f.build(backend='thread') as flow:
            flow.index()
            ...

    You can also use the shortcuts, e.g. :py:meth:`add_encoder`, :py:meth:`add_preprocessor`.

    It is recommend to use flow in the context manner as showed above.

    Note the different default copy behaviors in :py:meth:`.add()` and :py:meth:`.build()`:
    :py:meth:`.add()` always copy the flow by default, whereas :py:meth:`.build()` modify the flow in place.
    You can change this behavior by giving an argument `copy_flow=False`.

    """

    def __init__(self, with_frontend: bool = True, is_trained: bool = True, *args, **kwargs):
        """
        Create a new Flow object.

        :param with_frontend: adding frontend service to the flow
        :param is_trained: indicating whether this flow is trained or not. if set to False then :py:meth:`index`
                            and :py:meth:`query` can not be called before :py:meth:`train`
        :param kwargs: keyword-value arguments that will be shared by all services
        """
        super().__init__(*args, **kwargs)
        self.logger = set_logger(self.__class__.__name__)
        self._service_nodes = OrderedDict()
        self._service_edges = {}
        self._service_name_counter = {k: 0 for k in service_map.keys()}
        self._service_contexts = []
        self._last_changed_service = []
        self._common_kwargs = kwargs
        self._frontend = None
        self._client = None
        self._build_level = BuildLevel.EMPTY
        self._backend = None
        self._init_with_frontend = False
        self.is_trained = is_trained
        if with_frontend:
            self.add_frontend(copy_flow=False)
            self._init_with_frontend = True
        else:
            self.logger.warning('with_frontend is set to False, you need to add_frontend() by yourself')

    @build_required(BuildLevel.GRAPH)
    def to_k8s_yaml(self) -> str:
        raise NotImplementedError

    @build_required(BuildLevel.GRAPH)
    def to_shell_script(self) -> str:
        raise NotImplementedError

    @build_required(BuildLevel.GRAPH)
    def to_swarm_yaml(self, image: str = 'gnes/gnes:latest-alpine') -> str:
        """
        Generate the docker swarm YAML compose file

        :param image: the default GNES docker image
        :return: the generated YAML compose file
        """
        from ruamel.yaml import YAML, StringIO
        _yaml = YAML()
        swarm_yml = {'version': '3.4',
                     'services': {}}

        for k, v in self._service_nodes.items():
            defaults_kwargs, _ = service_map[v['service']]['parser']().parse_known_args(
                ['--yaml_path', 'TrainableBase'])
            non_default_kwargs = {k: v for k, v in vars(v['parsed_args']).items() if getattr(defaults_kwargs, k) != v}
            if not isinstance(non_default_kwargs.get('yaml_path', ''), str):
                non_default_kwargs['yaml_path'] = v['kwargs']['yaml_path']

            num_replicas = None
            if 'num_parallel' in non_default_kwargs:
                num_replicas = non_default_kwargs.pop('num_parallel')

            swarm_yml['services'][k] = {
                'image': v['kwargs'].get('image', image),
                'command': '%s %s' % (
                    service_map[v['service']]['cmd'],
                    ' '.join(['--%s %s' % (k, v) for k, v in non_default_kwargs.items()]))
            }
            if num_replicas and num_replicas > 1:
                swarm_yml['services'][k]['deploy'] = {'replicas': num_replicas}

        stream = StringIO()
        _yaml.dump(swarm_yml, stream)
        return stream.getvalue().strip()

    def to_python_code(self, indent: int = 4) -> str:
        """
        Generate the python code of this flow

        :param indent: the number of whitespaces of indent
        :return: the generated python code
        """
        py_code = ['from gnes.flow import Flow', '']
        kwargs = []
        if not self._init_with_frontend:
            kwargs.append('with_frontend=False')
        if self._common_kwargs:
            kwargs.extend('%s=%s' % (k, v) for k, v in self._common_kwargs.items())
        py_code.append('f = (Flow(%s)' % (', '.join(kwargs)))

        known_service = set()
        last_add_name = ''

        for k, v in self._service_nodes.items():
            kwargs = OrderedDict()
            kwargs['service'] = str(v['service'])
            kwargs['name'] = k
            kwargs['recv_from'] = '[%s]' % (
                ','.join({'\'%s\'' % k for k in v['incomes'] if k in known_service}))
            if kwargs['recv_from'] == '[\'%s\']' % last_add_name:
                kwargs.pop('recv_from')
            kwargs['send_to'] = '[%s]' % (','.join({'\'%s\'' % k for k in v['outgoings'] if k in known_service}))

            known_service.add(k)
            last_add_name = k

            py_code.append('%s.add(%s)' % (
                ' ' * indent,
                ', '.join(
                    '%s=%s' % (kk, '\'%s\'' % vv if isinstance(vv, str)
                                                    and not vv.startswith('\'') and not vv.startswith('[')
                    else vv) for kk, vv
                    in
                    (list(kwargs.items()) + list(v['kwargs'].items())) if
                    vv and vv != '[]' and kk not in self._common_kwargs)))

        py_code[-1] += ')'

        py_code.extend(['',
                        '# build the flow and visualize it',
                        'f.build(backend=None).to_url()'
                        ])
        py_code.extend(['',
                        '# use this flow in multi-thread mode for indexing',
                        'with f.build(backend=\'thread\') as fl:',
                        '%sfl.index(txt_file=\'test.txt\')' % (' ' * indent)
                        ])
        py_code.append('')

        return '\n'.join(py_code)

    @build_required(BuildLevel.GRAPH)
    def to_mermaid(self, left_right: bool = True) -> str:
        """
        Output the mermaid graph for visualization

        :param left_right: render the flow in left-to-right manner, otherwise top-down manner.
        :return: a mermaid-formatted string
        """

        # fill, stroke
        service_color = {
            Service.Frontend: ('#FFE0E0', '#000'),
            Service.Router: ('#C9E8D2', '#000'),
            Service.Encoder: ('#FFDAAF', '#000'),
            Service.Preprocessor: ('#CED7EF', '#000'),
            Service.Indexer: ('#FFFBC1', '#000'),
        }

        mermaid_graph = OrderedDict()
        cls_dict = defaultdict(set)
        replicas_dict = {}

        for k, v in self._service_nodes.items():
            mermaid_graph[k] = []
            num_replicas = getattr(v['parsed_args'], 'num_parallel', 1)
            if num_replicas > 1:
                head_router = k + '_HEAD'
                tail_router = k + '_TAIL'
                replicas_dict[k] = (head_router, tail_router)
                cls_dict[Service.Router].add(head_router)
                cls_dict[Service.Router].add(tail_router)
                p_r = '((%s))'
                k_service = v['service']
                p_e = '((%s))' if k_service == Service.Router else '(%s)'

                mermaid_graph[k].append('subgraph %s["%s (replias=%d)"]' % (k, k, num_replicas))
                for j in range(num_replicas):
                    r = k + '_%d' % j
                    cls_dict[k_service].add(r)
                    mermaid_graph[k].append('\t%s%s-->%s%s' % (head_router, p_r % 'router', r, p_e % r))
                    mermaid_graph[k].append('\t%s%s-->%s%s' % (r, p_e % r, tail_router, p_r % 'router'))
                mermaid_graph[k].append('end')
                mermaid_graph[k].append(
                    'style %s fill:%s,stroke:%s,stroke-width:2px,stroke-dasharray:5,stroke-opacity:0.3,fill-opacity:0.5' % (
                        k, service_color[k_service][0], service_color[k_service][1]))

        for k, ed_type in self._service_edges.items():
            start_node, end_node = k.split('-')
            cur_node = mermaid_graph[start_node]

            s_service = self._service_nodes[start_node]['service']
            e_service = self._service_nodes[end_node]['service']

            start_node_text = start_node
            end_node_text = end_node

            # check if is in replicas
            if start_node in replicas_dict:
                start_node = replicas_dict[start_node][1]  # outgoing
                s_service = Service.Router
                start_node_text = 'router'
            if end_node in replicas_dict:
                end_node = replicas_dict[end_node][0]  # incoming
                e_service = Service.Router
                end_node_text = 'router'

            # always plot frontend at the start and the end
            if e_service == Service.Frontend:
                end_node_text = end_node
                end_node += '_END'

            cls_dict[s_service].add(start_node)
            cls_dict[e_service].add(end_node)
            p_s = '((%s))' if s_service == Service.Router else '(%s)'
            p_e = '((%s))' if e_service == Service.Router else '(%s)'
            cur_node.append('\t%s%s-- %s -->%s%s' % (
                start_node, p_s % start_node_text, ed_type,
                end_node, p_e % end_node_text))

        style = ['classDef %sCLS fill:%s,stroke:%s,stroke-width:1px;' % (k, v[0], v[1]) for k, v in
                 service_color.items()]
        class_def = ['class %s %sCLS;' % (','.join(v), k) for k, v in cls_dict.items()]
        mermaid_str = '\n'.join(
            ['graph %s' % ('LR' if left_right else 'TD')] + [ss for s in mermaid_graph.values() for ss in
                                                             s] + style + class_def)

        return mermaid_str

    @build_required(BuildLevel.GRAPH)
    def to_url(self, **kwargs) -> str:
        """
        Rendering the current flow as a url points to a SVG, it needs internet connection

        :param kwargs: keyword arguments of :py:meth:`to_mermaid`
        :return: the url points to a SVG
        """
        import base64
        mermaid_str = self.to_mermaid(**kwargs)
        encoded_str = base64.b64encode(bytes(mermaid_str, 'utf-8')).decode('utf-8')
        return 'https://mermaidjs.github.io/mermaid-live-editor/#/view/%s' % encoded_str

    @build_required(BuildLevel.GRAPH)
    def to_jpg(self, path: str = 'flow.jpg', **kwargs) -> None:
        """
        Rendering the current flow as a jpg image, this will call :py:meth:`to_mermaid` and it needs internet connection

        :param path: the file path of the image
        :param kwargs: keyword arguments of :py:meth:`to_mermaid`
        :return:
        """

        from urllib.request import Request, urlopen
        encoded_str = self.to_url().replace('https://mermaidjs.github.io/mermaid-live-editor/#/view/', '')
        self.logger.warning('jpg exporting relies on https://mermaid.ink/, but it is not very stable. '
                            'some syntax are not supported, please use with caution.')
        self.logger.info('downloading as jpg...')
        req = Request('https://mermaid.ink/img/%s' % encoded_str, headers={'User-Agent': 'Mozilla/5.0'})
        with open(path, 'wb') as fp:
            fp.write(urlopen(req).read())
        self.logger.info('done')

    def train(self, bytes_gen: Iterator[bytes] = None, **kwargs):
        """Do training on the current flow

        It will start a :py:class:`CLIClient` and call :py:func:`train`.

        :param bytes_gen: An iterator of bytes. If not given, then you have to specify it in `kwargs`.
        :param kwargs: accepts all keyword arguments of `gnes client` CLI
        """
        self._call_client(bytes_gen, mode='train', **kwargs)

    def index(self, bytes_gen: Iterator[bytes] = None, **kwargs):
        """Do indexing on the current flow

        It will start a :py:class:`CLIClient` and call :py:func:`index`.

        :param bytes_gen: An iterator of bytes. If not given, then you have to specify it in `kwargs`.
        :param kwargs: accepts all keyword arguments of `gnes client` CLI
        """
        self._call_client(bytes_gen, mode='index', **kwargs)

    def query(self, bytes_gen: Iterator[bytes] = None, **kwargs):
        """Do indexing on the current flow

        It will start a :py:class:`CLIClient` and call :py:func:`query`.

        :param bytes_gen: An iterator of bytes. If not given, then you have to specify it in `kwargs`.
        :param kwargs: accepts all keyword arguments of `gnes client` CLI
        """
        self._call_client(bytes_gen, mode='query', **kwargs)

    @build_required(BuildLevel.RUNTIME)
    def _call_client(self, bytes_gen: Iterator[bytes] = None, **kwargs):
        from ..cli.parser import set_client_cli_parser
        from ..client.cli import CLIClient

        args, p_args, unk_args = self._get_parsed_args(self, set_client_cli_parser, kwargs)
        p_args.grpc_port = self._service_nodes[self._frontend]['parsed_args'].grpc_port
        p_args.grpc_host = self._service_nodes[self._frontend]['parsed_args'].grpc_host
        c = CLIClient(p_args, start_at_init=False)
        if bytes_gen:
            c.bytes_generator = bytes_gen
        c.start()

    def add_frontend(self, *args, **kwargs) -> 'Flow':
        """Add a frontend to the current flow, a shortcut of :py:meth:`add(Service.Frontend)`.
        Usually you dont need to call this function explicitly, a flow object contains a frontend service by default.
        This function is useful when you build a flow without the frontend and want to customize the frontend later.
        """
        return self.add(Service.Frontend, *args, **kwargs)

    def add_encoder(self, *args, **kwargs) -> 'Flow':
        """Add an encoder to the current flow, a shortcut of :py:meth:`add(Service.Encoder)`"""
        return self.add(Service.Encoder, *args, **kwargs)

    def add_indexer(self, *args, **kwargs) -> 'Flow':
        """Add an indexer to the current flow, a shortcut of :py:meth:`add(Service.Indexer)`"""
        return self.add(Service.Indexer, *args, **kwargs)

    def add_preprocessor(self, *args, **kwargs) -> 'Flow':
        """Add a preprocessor to the current flow, a shortcut of :py:meth:`add(Service.Preprocessor)`"""
        return self.add(Service.Preprocessor, *args, **kwargs)

    def add_router(self, *args, **kwargs) -> 'Flow':
        """Add a router to the current flow, a shortcut of :py:meth:`add(Service.Router)`"""
        return self.add(Service.Router, *args, **kwargs)

    def set_last_service(self, name: str, copy_flow: bool = True) -> 'Flow':
        """
        Set a service as the last service in the flow, useful when modifying the flow.

        :param name: the name of the existing service
        :param copy_flow: when set to true, then always copy the current flow and do the modification on top of it then return, otherwise, do in-line modification
        :return: a (new) flow object with modification
        """
        op_flow = copy.deepcopy(self) if copy_flow else self

        if name not in op_flow._service_nodes:
            raise FlowMissingNode('recv_from: %s can not be found in this Flow' % name)

        if op_flow._last_changed_service and name == op_flow._last_changed_service[-1]:
            pass
        else:
            op_flow._last_changed_service.append(name)

        # graph is now changed so we need to
        # reset the build level to the lowest
        op_flow._build_level = BuildLevel.EMPTY

        return op_flow

    def set(self, name: str, recv_from: Union[str, Tuple[str], List[str], 'Service'] = None,
            send_to: Union[str, Tuple[str], List[str], 'Service'] = None,
            copy_flow: bool = True,
            clear_old_attr: bool = False,
            as_last_service: bool = False,
            **kwargs) -> 'Flow':
        """
        Set the attribute of an existing service (added by :py:meth:`add`) in the flow.
        For the attributes or kwargs that aren't given, they will remain unchanged as before.

        :param name: the name of the existing service
        :param recv_from: the name of the service(s) that this service receives data from.
                           One can also use 'Service.Frontend' to indicate the connection with the frontend.
        :param send_to:  the name of the service(s) that this service sends data to.
                           One can also use 'Service.Frontend' to indicate the connection with the frontend.
        :param copy_flow: when set to true, then always copy the current flow and do the modification on top of it then return, otherwise, do in-line modification
        :param clear_old_attr: remove old attribute value before setting the new one
        :param as_last_service: whether setting the changed service as the last service in the graph
        :param kwargs: other keyword-value arguments that the service CLI supports
        :return: a (new) flow object with modification
        """
        op_flow = copy.deepcopy(self) if copy_flow else self

        if name not in op_flow._service_nodes:
            raise FlowMissingNode('recv_from: %s can not be found in this Flow' % name)

        node = op_flow._service_nodes[name]
        service = node['service']

        if recv_from:
            recv_from = op_flow._parse_service_endpoints(op_flow, name, recv_from, connect_to_last_service=True)

            if clear_old_attr:
                node['incomes'] = recv_from
                # remove all edges point to this service
                for n in op_flow._service_nodes.values():
                    if name in n['outgoings']:
                        n['outgoings'].remove(name)
            else:
                node['incomes'] = node['incomes'].union(recv_from)

            # add it the new edge back
            for s in recv_from:
                op_flow._service_nodes[s]['outgoings'].add(name)

        if send_to:
            send_to = op_flow._parse_service_endpoints(op_flow, name, send_to, connect_to_last_service=False)
            node['outgoings'] = send_to
            if clear_old_attr:
                # remove all edges this service point to
                for n in op_flow._service_nodes.values():
                    if name in n['incomes']:
                        n['incomes'].remove(name)
            else:
                node['outgoings'] = node['outgoings'].union(send_to)

            for s in send_to:
                op_flow._service_nodes[s]['incomes'].add(name)

        if kwargs:
            if not clear_old_attr:
                node['kwargs'].update(kwargs)
                kwargs = node['kwargs']
            args, p_args, unk_args = op_flow._get_parsed_args(op_flow, service_map[service]['parser'], kwargs)
            node.update({
                'args': args,
                'parsed_args': p_args,
                'kwargs': kwargs,
                'unk_args': unk_args
            })

        if as_last_service:
            op_flow.set_last_service(name, False)

        # graph is now changed so we need to
        # reset the build level to the lowest
        op_flow._build_level = BuildLevel.EMPTY

        return op_flow

    def remove(self, name: str = None, copy_flow: bool = True) -> 'Flow':
        """
        Remove a service from the flow.

        :param name: the name of the existing service
        :param copy_flow: when set to true, then always copy the current flow and do the modification on top of it then return, otherwise, do in-line modification
        :return: a (new) flow object with modification
        """

        op_flow = copy.deepcopy(self) if copy_flow else self

        if name not in op_flow._service_nodes:
            raise FlowMissingNode('recv_from: %s can not be found in this Flow' % name)

        op_flow._service_nodes.pop(name)

        # remove all edges point to this service
        for n in op_flow._service_nodes.values():
            if name in n['outgoings']:
                n['outgoings'].remove(name)
            if name in n['incomes']:
                n['incomes'].remove(name)

        if op_flow._service_nodes:
            op_flow._last_changed_service = [v for v in op_flow._last_changed_service if v != name]
        else:
            op_flow._last_changed_service = []

        # graph is now changed so we need to
        # reset the build level to the lowest
        op_flow._build_level = BuildLevel.EMPTY

        return op_flow

    def add(self, service: Union['Service', str],
            name: str = None,
            recv_from: Union[str, Tuple[str], List[str], 'Service'] = None,
            send_to: Union[str, Tuple[str], List[str], 'Service'] = None,
            copy_flow: bool = True,
            **kwargs) -> 'Flow':
        """
        Add a service to the current flow object and return the new modified flow object.
        The attribute of the service can be later changed with :py:meth:`set` or deleted with :py:meth:`remove`

        :param service: a 'Service' enum or string, possible choices: Encoder, Router, Preprocessor, Indexer, Frontend
        :param name: the name identifier of the service, can be used in 'recv_from',
                    'send_to', :py:meth:`set` and :py:meth:`remove`.
        :param recv_from: the name of the service(s) that this service receives data from.
                           One can also use 'Service.Frontend' to indicate the connection with the frontend.
        :param send_to:  the name of the service(s) that this service sends data to.
                           One can also use 'Service.Frontend' to indicate the connection with the frontend.
        :param copy_flow: when set to true, then always copy the current flow and do the modification on top of it then return, otherwise, do in-line modification
        :param kwargs: other keyword-value arguments that the service CLI supports
        :return: a (new) flow object with modification
        """

        op_flow = copy.deepcopy(self) if copy_flow else self

        if isinstance(service, str):
            service = Service.from_string(service)

        if service not in service_map:
            raise ValueError('service: %s is not supported, should be one of %s' % (service, service_map.keys()))

        if name in op_flow._service_nodes:
            raise FlowTopologyError('name: %s is used in this Flow already!' % name)
        if not name:
            name = '%s%d' % (service, op_flow._service_name_counter[service])
            op_flow._service_name_counter[service] += 1
        if not name.isidentifier():
            raise ValueError('name: %s is invalid, please follow the python variable name conventions' % name)

        if service == Service.Frontend:
            if op_flow._frontend:
                raise FlowTopologyError('frontend is already in this Flow')
            op_flow._frontend = name

        recv_from = op_flow._parse_service_endpoints(op_flow, name, recv_from, connect_to_last_service=True)
        send_to = op_flow._parse_service_endpoints(op_flow, name, send_to, connect_to_last_service=False)

        args, p_args, unk_args = op_flow._get_parsed_args(op_flow, service_map[service]['parser'], kwargs)

        op_flow._service_nodes[name] = {
            'service': service,
            'parsed_args': p_args,
            'args': args,
            'incomes': recv_from,
            'outgoings': send_to,
            'kwargs': kwargs,
            'unk_args': unk_args
        }

        # direct all income services' output to the current service
        for s in recv_from:
            op_flow._service_nodes[s]['outgoings'].add(name)
        for s in send_to:
            op_flow._service_nodes[s]['incomes'].add(name)

        op_flow.set_last_service(name, False)

        # graph is now changed so we need to
        # reset the build level to the lowest
        op_flow._build_level = BuildLevel.EMPTY

        return op_flow

    @staticmethod
    def _parse_service_endpoints(op_flow, cur_service_name, service_endpoint, connect_to_last_service=False):
        # parsing recv_from
        if isinstance(service_endpoint, str):
            service_endpoint = [service_endpoint]
        elif service_endpoint == Service.Frontend:
            service_endpoint = [op_flow._frontend]
        elif not service_endpoint:
            if op_flow._last_changed_service and connect_to_last_service:
                service_endpoint = [op_flow._last_changed_service[-1]]
            else:
                service_endpoint = []
        if isinstance(service_endpoint, list) or isinstance(service_endpoint, tuple):
            for s in service_endpoint:
                if s == cur_service_name:
                    raise FlowTopologyError('the income of a service can not be itself')
                if s not in op_flow._service_nodes:
                    raise FlowMissingNode('recv_from: %s can not be found in this Flow' % s)
        else:
            raise ValueError('recv_from=%s is not parsable' % service_endpoint)
        return set(service_endpoint)

    @staticmethod
    def _get_parsed_args(op_flow, service_arg_parser, kwargs):
        kwargs.update(op_flow._common_kwargs)
        args = []
        for k, v in kwargs.items():
            if isinstance(v, bool):
                if v:
                    if not k.startswith('no_') and not k.startswith('no-'):
                        args.append('--%s' % k)
                    else:
                        args.append('--%s' % k[3:])
                else:
                    if k.startswith('no_') or k.startswith('no-'):
                        args.append('--%s' % k)
                    else:
                        args.append('--no_%s' % k)
            else:
                args.extend(['--%s' % k, str(v)])
        try:
            p_args, unknown_args = service_arg_parser().parse_known_args(args)
            if unknown_args:
                op_flow.logger.warning('not sure what these arguments are: %s' % unknown_args)
        except SystemExit:
            raise ValueError('bad arguments for service "%s", '
                             'you may want to double check your args "%s"' % (service_arg_parser, args))
        return args, p_args, unknown_args

    def _build_graph(self, copy_flow: bool) -> 'Flow':
        op_flow = copy.deepcopy(self) if copy_flow else self

        op_flow._service_edges.clear()

        if not op_flow._frontend:
            raise FlowIncompleteError('frontend does not exist, you may need to add_frontend()')

        if not op_flow._last_changed_service or not op_flow._service_nodes:
            raise FlowTopologyError('flow is empty?')

        # close the loop
        op_flow._service_nodes[op_flow._frontend]['incomes'] = {op_flow._last_changed_service[-1]}

        # build all edges
        for k, v in op_flow._service_nodes.items():
            for s in v['incomes']:
                op_flow._service_edges['%s-%s' % (s, k)] = ''
            for t in v['outgoings']:
                op_flow._service_edges['%s-%s' % (k, t)] = ''

        for k in op_flow._service_edges.keys():
            start_node, end_node = k.split('-')
            edges_with_same_start = [ed for ed in op_flow._service_edges.keys() if ed.startswith(start_node)]
            edges_with_same_end = [ed for ed in op_flow._service_edges.keys() if ed.endswith(end_node)]

            s_pargs = op_flow._service_nodes[start_node]['parsed_args']
            e_pargs = op_flow._service_nodes[end_node]['parsed_args']

            # Rule
            # if a node has multiple income/outgoing services,
            # then its socket_in/out must be PULL_BIND or PUB_BIND
            # otherwise it should be different than its income
            # i.e. income=BIND => this=CONNECT, income=CONNECT => this = BIND
            #
            # when a socket is BIND, then host must NOT be set, aka default host 0.0.0.0
            # host_in and host_out is only set when corresponding socket is CONNECT

            if len(edges_with_same_start) > 1 and len(edges_with_same_end) == 1:
                s_pargs.socket_out = SocketType.PUB_BIND
                s_pargs.host_out = BaseService.default_host
                e_pargs.socket_in = SocketType.SUB_CONNECT
                e_pargs.host_in = start_node
                e_pargs.port_in = s_pargs.port_out
                op_flow._service_edges[k] = 'PUB-sub'
            elif len(edges_with_same_end) > 1 and len(edges_with_same_start) == 1:
                s_pargs.socket_out = SocketType.PUSH_CONNECT
                s_pargs.host_out = end_node
                e_pargs.socket_in = SocketType.PULL_BIND
                e_pargs.host_in = BaseService.default_host
                s_pargs.port_out = e_pargs.port_in
                op_flow._service_edges[k] = 'push-PULL'
            elif len(edges_with_same_start) == 1 and len(edges_with_same_end) == 1:
                # in this case, either side can be BIND
                # we prefer frontend to be always BIND
                # check if either node is frontend
                if start_node == op_flow._frontend:
                    s_pargs.socket_out = SocketType.PUSH_BIND
                    e_pargs.socket_in = SocketType.PULL_CONNECT
                elif end_node == op_flow._frontend:
                    s_pargs.socket_out = SocketType.PUSH_CONNECT
                    e_pargs.socket_in = SocketType.PULL_BIND
                else:
                    e_pargs.socket_in = s_pargs.socket_out.paired

                if s_pargs.socket_out.is_bind:
                    s_pargs.host_out = BaseService.default_host
                    e_pargs.host_in = start_node
                    e_pargs.port_in = s_pargs.port_out
                    op_flow._service_edges[k] = 'PUSH-pull'
                elif e_pargs.socket_in.is_bind:
                    s_pargs.host_out = end_node
                    e_pargs.host_in = BaseService.default_host
                    s_pargs.port_out = e_pargs.port_in
                    op_flow._service_edges[k] = 'push-PULL'
                else:
                    raise FlowTopologyError('edge %s -> %s is ambiguous, at least one socket should be BIND')
            else:
                raise FlowTopologyError('found %d edges start with %s and %d edges end with %s, '
                                        'this type of topology is ambiguous and should not exist, '
                                        'i can not determine the socket type' % (
                                            len(edges_with_same_start), start_node, len(edges_with_same_end), end_node))

        op_flow._build_level = BuildLevel.GRAPH
        return op_flow

    def build(self, backend: Optional[str] = 'thread', copy_flow: bool = False, *args, **kwargs) -> 'Flow':
        """
        Build the current flow and make it ready to use

        :param backend: supported 'thread', 'process', 'swarm', 'k8s', 'shell', if None then only build graph only
        :param copy_flow: return the copy of the current flow
        :return: the current flow (by default)
        """

        op_flow = self._build_graph(copy_flow)

        if not backend:
            op_flow.logger.warning('no specified backend, build_level stays at %s, '
                                   'and you can not run this flow.' % op_flow._build_level)
        elif backend in {'thread', 'process'}:
            op_flow._service_contexts.clear()
            for v in op_flow._service_nodes.values():
                p_args = v['parsed_args']
                p_args.parallel_backend = backend
                # for thread and process backend which runs locally, host_in and host_out should not be set
                p_args.host_in = BaseService.default_host
                p_args.host_out = BaseService.default_host
                op_flow._service_contexts.append((service_map[v['service']]['builder'], p_args))
            op_flow._build_level = BuildLevel.RUNTIME
        else:
            raise NotImplementedError('backend=%s is not supported yet' % backend)

        return op_flow

    def __call__(self, *args, **kwargs):
        return self.build(*args, **kwargs)

    def __enter__(self):
        if self._build_level.value < BuildLevel.RUNTIME.value:
            self.logger.warning(
                'current build_level=%s, lower than required. '
                'build the flow now via build() with default parameters' % self._build_level)
            self.build(copy_flow=False)
        self._service_stack = ExitStack()
        for k, v in self._service_contexts:
            self._service_stack.enter_context(k(v))

        self.logger.critical('flow is built and ready, current build level is %s' % self._build_level)
        return self

    def close(self):
        if hasattr(self, '_service_stack'):
            self._service_stack.close()
        self._build_level = BuildLevel.EMPTY
        self.logger.critical(
            'flow is closed and all resources should be released already, current build level is %s' % self._build_level)

    def __eq__(self, other):
        """
        Comparing the topology of a flow with another flow.
        Identification is defined by whether two flows share the same set of edges.

        :param other: the second flow object
        :return:
        """

        if self._build_level.value < BuildLevel.GRAPH.value:
            a = self.build(backend=None, copy_flow=True)
        else:
            a = self

        if other._build_level.value < BuildLevel.GRAPH.value:
            b = other.build(backend=None, copy_flow=True)
        else:
            b = other

        return a._service_edges == b._service_edges



from functools import wraps

from ..cli.parser import set_router_parser, set_indexer_parser, \
    set_frontend_parser, set_preprocessor_parser, \
    set_encoder_parser
from ..service.base import BetterEnum, ServiceManager
from ..service.encoder import EncoderService
from ..service.frontend import FrontendService
from ..service.indexer import IndexerService
from ..service.preprocessor import PreprocessorService
from ..service.router import RouterService


class BuildLevel(BetterEnum):
    EMPTY = 0
    GRAPH = 1
    RUNTIME = 2


class Service(BetterEnum):
    Frontend = 0
    Encoder = 1
    Router = 2
    Indexer = 3
    Preprocessor = 4


class FlowIncompleteError(ValueError):
    """Exception when the flow missing some important component to run"""


class FlowTopologyError(ValueError):
    """Exception when the topology is ambiguous"""


class FlowMissingNode(ValueError):
    """Exception when the topology is ambiguous"""


class FlowBuildLevelMismatch(ValueError):
    """Exception when required level is higher than the current build level"""


def build_required(required_level: 'BuildLevel'):
    def __build_level(func):
        @wraps(func)
        def arg_wrapper(self, *args, **kwargs):
            if hasattr(self, '_build_level'):
                if self._build_level.value >= required_level.value:
                    return func(self, *args, **kwargs)
                else:
                    raise FlowBuildLevelMismatch(
                        'build_level check failed for %r, required level: %s, actual level: %s' % (
                            func, required_level, self._build_level))
            else:
                raise AttributeError('%r has no attribute "_build_level"' % self)

        return arg_wrapper

    return __build_level


service_map = {
    Service.Encoder: {
        'parser': set_encoder_parser,
        'builder': lambda x: ServiceManager(EncoderService, x),
        'cmd': 'encode'},
    Service.Router: {
        'parser': set_router_parser,
        'builder': lambda x: ServiceManager(RouterService, x),
        'cmd': 'route',
    },
    Service.Indexer: {
        'parser': set_indexer_parser,
        'builder': lambda x: ServiceManager(IndexerService, x),
        'cmd': 'index'
    },
    Service.Frontend: {
        'parser': set_frontend_parser,
        'builder': FrontendService,
        'cmd': 'frontend'
    },
    Service.Preprocessor: {
        'parser': set_preprocessor_parser,
        'builder': lambda x: ServiceManager(PreprocessorService, x),
        'cmd': 'preprocess'
    }
}



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

def encode_video(images: List['np.ndarray'], frame_rate: int, pix_fmt: str = 'rgb24'):
    import webp
    from PIL import Image

    height, width, _ = images[0].shape
    if pix_fmt == 'rgb24':
        pix_fmt = 'RGB'
    # Save an animation
    enc = webp.WebPAnimEncoder.new(width, height)
    timestamp_ms = 0
    duration = 1000 // frame_rate
    for x in images:
        img = Image.fromarray(x.copy(), pix_fmt)
        pic = webp.WebPPicture.from_pil(img)
        enc.encode_frame(pic, timestamp_ms)
        timestamp_ms += duration
    anim_data = enc.assemble(timestamp_ms)
    return bytes(anim_data.buffer())



import copy
import os
import unittest

import numpy as np
from gnes.proto import gnes_pb2, array2blob

from gnes.preprocessor.base import BaseVideoPreprocessor


class TestVideoEncoder(unittest.TestCase):
    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.mp4_yaml_path = os.path.join(self.dirname, 'yaml', 'mp4-encoder.yml')
        self.gif_yaml_path = os.path.join(self.dirname, 'yaml', 'gif-encoder.yml')
        self.dump_path = os.path.join(self.dirname, 'video_encoder.bin')
        self.frames_path = os.path.join(self.dirname, 'frames', 'frames.npy')
        self.mp4_encoder = BaseVideoPreprocessor.load_yaml(self.mp4_yaml_path)
        self.gif_encoder = BaseVideoPreprocessor.load_yaml(self.gif_yaml_path)
        self.video_frames = np.load(self.frames_path)


    def test_mp4_encoder(self):
        raw_data = array2blob(self.video_frames)

        doc = gnes_pb2.Document()
        doc.doc_type = gnes_pb2.Document.VIDEO
        doc.raw_video.CopyFrom(raw_data)
        self.mp4_encoder.apply(doc)
        doc1 = copy.deepcopy(doc)

        doc = gnes_pb2.Document()
        doc.doc_type = gnes_pb2.Document.VIDEO
        chunk = doc.chunks.add()
        chunk.blob.CopyFrom(raw_data)
        self.mp4_encoder.apply(doc)
        doc2 = copy.deepcopy(doc)

        self.assertEqual(doc1.raw_bytes, doc2.chunks[0].raw)

    def test_gif_encoder(self):
        raw_data = array2blob(self.video_frames)

        doc = gnes_pb2.Document()
        doc.doc_type = gnes_pb2.Document.VIDEO
        doc.raw_video.CopyFrom(raw_data)
        self.gif_encoder.apply(doc)
        doc1 = copy.deepcopy(doc)

        doc = gnes_pb2.Document()
        doc.doc_type = gnes_pb2.Document.VIDEO
        chunk = doc.chunks.add()
        chunk.blob.CopyFrom(raw_data)
        self.gif_encoder.apply(doc)
        doc2 = copy.deepcopy(doc)

        self.assertEqual(doc1.raw_bytes, doc2.chunks[0].raw)

    def test_empty_doc(self):
        doc = gnes_pb2.Document()
        doc.doc_type = gnes_pb2.Document.VIDEO
        self.mp4_encoder.apply(doc)

    def test_dump_load(self):
        raw_data = array2blob(self.video_frames)

        doc = gnes_pb2.Document()
        doc.doc_type = gnes_pb2.Document.VIDEO
        doc.raw_video.CopyFrom(raw_data)
        self.mp4_encoder.apply(doc)
        doc1 = copy.deepcopy(doc)

        self.mp4_encoder.dump(self.dump_path)

        encoder = BaseVideoPreprocessor.load(self.dump_path)

        doc = gnes_pb2.Document()
        doc.doc_type = gnes_pb2.Document.VIDEO
        chunk = doc.chunks.add()
        chunk.blob.CopyFrom(raw_data)
        encoder.apply(doc)
        doc2 = copy.deepcopy(doc)

        self.assertEqual(doc1.raw_bytes, doc2.chunks[0].raw)


    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


def _start_service(cls, args):
    from ..service.base import ServiceManager
    with ServiceManager(cls, args) as es:
        es.join()


def grpc(args):
    from ..service.grpc import GRPCService
    _start_service(GRPCService, args)


def preprocess(args):
    from ..service.preprocessor import PreprocessorService
    _start_service(PreprocessorService, args)


def encode(args):
    from ..service.encoder import EncoderService
    _start_service(EncoderService, args)


def index(args):
    from ..service.indexer import IndexerService
    _start_service(IndexerService, args)


def route(args):
    from ..service.router import RouterService
    _start_service(RouterService, args)


def frontend(args):
    from ..service.frontend import FrontendService
    _start_service(FrontendService, args)


def client(args):
    if args.client == 'http':
        return _client_http(args)
    elif args.client == 'cli':
        return _client_cli(args)
    else:
        raise ValueError('gnes client must follow with a client type from {http, cli, benchmark...}\n'
                         'see "gnes client --help" for details')


def healthcheck(args):
    from ..service.base import send_ctrl_message
    from ..proto import gnes_pb2, add_version
    import time
    ctrl_addr = 'tcp://%s:%d' % (args.host, args.port)
    msg = gnes_pb2.Message()
    add_version(msg.envelope)
    msg.request.control.command = gnes_pb2.Request.ControlRequest.STATUS
    for j in range(args.retries):
        r = send_ctrl_message(ctrl_addr, msg, timeout=args.timeout)
        if not r:
            print('%s is not responding, retry (%d/%d) in 1s' % (ctrl_addr, j + 1, args.retries))
        else:
            print('%s returns %s' % (ctrl_addr, r))
            exit(0)
        time.sleep(1)
    exit(1)


def _client_http(args):
    from ..client.http import HttpClient
    HttpClient(args).start()


def _client_cli(args):
    from ..client.cli import CLIClient
    CLIClient(args)


def compose(args):
    from ..composer.base import YamlComposer
    from ..composer.flask import YamlComposerFlask
    from ..composer.http import YamlComposerHttp

    if args.flask:
        YamlComposerFlask(args).run()
    elif args.serve:
        YamlComposerHttp(args).run()
    else:
        YamlComposer(args).build_all()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
import math

from gnes.preprocessor.base import BaseVideoPreprocessor
from gnes.proto import gnes_pb2, array2blob, blob2array


class FrameSelectPreprocessor(BaseVideoPreprocessor):

    def __init__(self,
                 sframes: int = 1,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.sframes = sframes

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)
        if len(doc.chunks) > 0:
            for chunk in doc.chunks:
                images = blob2array(chunk.blob)
                if len(images) == 0:
                    self.logger.warning("this chunk has no frame!")
                elif self.sframes == 1:
                    idx = [int(len(images) / 2)]
                    chunk.blob.CopyFrom(array2blob(images[idx]))
                elif self.sframes > 0 and len(images) > self.sframes:
                    if len(images) >= 2 * self.sframes:
                        step = math.ceil(len(images) / self.sframes)
                        chunk.blob.CopyFrom(array2blob(images[::step]))
                    else:
                        idx = np.sort(np.random.choice(len(images), self.sframes, replace=False))
                        chunk.blob.CopyFrom(array2blob(images[idx]))
                del images
        else:
            self.logger.error(
                'bad document: "doc.chunks" is empty!')



import copy
import random
import unittest

import numpy as np

from gnes.cli.parser import _set_client_parser
from gnes.client.base import ZmqClient
from gnes.helper import TimeContext
from gnes.proto import gnes_pb2, array2blob, blob2array
from gnes.service.base import SocketType


class TestSqueezedSendRecv(unittest.TestCase):
    def setUp(self):
        self.c1_args = _set_client_parser().parse_args([
            '--port_in', str(5678),
            '--port_out', str(5679),
            '--socket_out', str(SocketType.PUSH_BIND),
            '--no-check_version'
        ])
        self.c2_args = _set_client_parser().parse_args([
            '--port_in', str(self.c1_args.port_out),
            '--port_out', str(self.c1_args.port_in),
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--no-check_version'
        ])

    def test_send_recv(self):
        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            msg = gnes_pb2.Message()
            msg.envelope.client_id = c1.args.identity
            d = msg.request.index.docs.add()
            d.raw_bytes = b'aa'
            c1.send_message(msg)
            r_msg = c2.recv_message()
            self.assertEqual(r_msg.request.index.docs[0].raw_bytes, d.raw_bytes)

    def test_send_recv_raw_bytes(self):
        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            msg = gnes_pb2.Message()
            msg.envelope.client_id = c1.args.identity
            for j in range(random.randint(10, 20)):
                d = msg.request.index.docs.add()
                d.raw_bytes = b'a' * random.randint(100, 1000)
            raw_bytes = copy.deepcopy([d.raw_bytes for d in msg.request.index.docs])
            c1.send_message(msg, squeeze_pb=True)
            r_msg = c2.recv_message()
            for d, o_d, r_d in zip(msg.request.index.docs, raw_bytes, r_msg.request.index.docs):
                self.assertEqual(d.raw_bytes, b'')
                self.assertEqual(o_d, r_d.raw_bytes)
                print('.', end='')
            print('checked %d docs' % len(msg.request.index.docs))

    def test_send_recv_response(self):
        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            msg = gnes_pb2.Message()
            msg.envelope.client_id = c1.args.identity
            msg.response.train.status = 2
            c1.send_message(msg, squeeze_pb=True)
            r_msg = c2.recv_message()
            self.assertEqual(msg.response.train.status, r_msg.response.train.status)

    def build_msgs(self):
        all_msgs = []
        num_msg = 20
        for j in range(num_msg):
            msg = gnes_pb2.Message()
            msg.envelope.client_id = 'abc'
            for j in range(random.randint(10, 20)):
                d = msg.request.index.docs.add()
                # each doc is about 1MB to 10MB
                d.raw_bytes = b'a' * random.randint(1000000, 10000000)
            all_msgs.append(msg)
        return all_msgs

    def build_msgs2(self, seed=0):
        all_msgs = []
        num_msg = 20
        random.seed(seed)
        np.random.seed(seed)
        for j in range(num_msg):
            msg = gnes_pb2.Message()
            msg.envelope.client_id = 'abc'
            for _ in range(random.randint(10, 20)):
                d = msg.request.index.docs.add()
                # each doc is about 1MB to 10MB
                for _ in range(random.randint(10, 20)):
                    c = d.chunks.add()
                    c.embedding.CopyFrom(array2blob(np.random.random([10, 20, 30])))
                    c.blob.CopyFrom(array2blob(np.random.random([10, 20, 30])))
            all_msgs.append(msg)
        return all_msgs

    def test_benchmark(self):
        all_msgs = self.build_msgs()

        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            with TimeContext('send, squeeze_pb=False'):
                for m in all_msgs:
                    c1.send_message(m)
            with TimeContext('recv, squeeze_pb=False'):
                for _ in all_msgs:
                    c2.recv_message()

        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            with TimeContext('send, squeeze_pb=True'):
                for m in all_msgs:
                    c1.send_message(m, squeeze_pb=True)
            with TimeContext('recv, squeeze_pb=True'):
                for _ in all_msgs:
                    c2.recv_message()

    def test_benchmark2(self):
        all_msgs = self.build_msgs()

        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            with TimeContext('send->recv, squeeze_pb=False'):
                for m in all_msgs:
                    c1.send_message(m, squeeze_pb=False)
                    c2.recv_message()

        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            with TimeContext('send->recv, squeeze_pb=True'):
                for m in all_msgs:
                    c1.send_message(m, squeeze_pb=True)
                    c2.recv_message()

    def test_benchmark3(self):
        all_msgs = self.build_msgs()
        all_msgs_bak = copy.deepcopy(all_msgs)

        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            for m, m1 in zip(all_msgs, all_msgs_bak):
                c1.send_message(m, squeeze_pb=True)
                r_m = c2.recv_message()
                for d, o_d, r_d in zip(m.request.index.docs, m1.request.index.docs, r_m.request.index.docs):
                    self.assertEqual(d.raw_bytes, b'')
                    self.assertEqual(o_d.raw_bytes, r_d.raw_bytes)

    def test_benchmark4(self):
        all_msgs = self.build_msgs2()

        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            with TimeContext('send->recv, squeeze_pb=False'):
                for m in all_msgs:
                    c1.send_message(m, squeeze_pb=False)
                    c2.recv_message()

        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            with TimeContext('send->recv, squeeze_pb=True'):
                for m in all_msgs:
                    c1.send_message(m, squeeze_pb=True)
                    c2.recv_message()

    def test_benchmark5(self):
        all_msgs = self.build_msgs2()
        all_msgs_bak = copy.deepcopy(all_msgs)

        with ZmqClient(self.c1_args) as c1, ZmqClient(self.c2_args) as c2:
            with TimeContext('send->recv, squeeze_pb=True'):
                for m, m1 in zip(all_msgs, all_msgs_bak):
                    c1.send_message(m, squeeze_pb=True)
                    r_m = c2.recv_message()

                    for d, r_d in zip(m1.request.index.docs, r_m.request.index.docs):
                        for c, r_c in zip(d.chunks, r_d.chunks):
                            np.allclose(blob2array(c.embedding), blob2array(r_c.embedding))
                            np.allclose(blob2array(c.blob), blob2array(r_c.blob))


from . import Flow

# these base flow define common topology in real world
# however, they are not usable directly unless you set the correct `yaml_path`

BaseIndexFlow = (Flow()
                 .add_preprocessor(name='prep', yaml_path='BasePreprocessor')
                 .add_encoder(name='enc', yaml_path='BaseEncoder')
                 .add_indexer(name='vec_idx', yaml_path='BaseIndexer')
                 .add_indexer(name='doc_idx', yaml_path='BaseIndexer', recv_from='prep')
                 .add_router(name='sync', yaml_path='BaseReduceRouter',
                             num_part=2, recv_from=['vec_idx', 'doc_idx']))

BaseQueryFlow = (Flow()
                 .add_preprocessor(name='prep', yaml_path='BasePreprocessor')
                 .add_encoder(name='enc', yaml_path='BaseEncoder')
                 .add_indexer(name='vec_idx', yaml_path='BaseIndexer')
                 .add_router(name='scorer', yaml_path='Chunk2DocTopkReducer')
                 .add_indexer(name='doc_idx', yaml_path='BaseIndexer'))



import os
import unittest
import numpy as np
from gnes.encoder.numeric.vlad import VladEncoder


class TestVladEncoder(unittest.TestCase):
    def setUp(self):
        self.mock_train_data = np.random.random([1, 200, 128]).astype(np.float32)
        self.mock_eval_data = np.random.random([2, 2, 128]).astype(np.float32)
        self.dump_path = os.path.join(os.path.dirname(__file__), 'vlad.bin')

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)

    def test_vlad_train(self):
        model = VladEncoder(20)
        model.train(self.mock_train_data)
        self.assertEqual(model.centroids.shape, (20, 128))
        v = model.encode(self.mock_eval_data)
        self.assertEqual(v.shape, (2, 2560))

    def test_vlad_dump_load(self):
        model = VladEncoder(20)
        model.train(self.mock_train_data)
        model.dump(self.dump_path)
        model_new = VladEncoder.load(self.dump_path)
        self.assertEqual(model_new.centroids.shape, (20, 128))



import threading
import time
from datetime import datetime

from . import helper


@helper.Singleton
class BaseIDGenerator(object):
    """
    Thread-safe (auto incremental) uuid generator
    """

    def __init__(self, start_id: int = 0, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self._lock = threading.Lock()
        self._next_id = start_id

    def reset(self, start_id: int = 0):
        with self._lock:
            self._next_id = start_id

    def next(self) -> int:
        with self._lock:
            temp = self._next_id
            self._next_id += 1
            return temp


@helper.Singleton
class SnowflakeIDGenerator(object):

    def __init__(self,
                 machine_id: int = 0,
                 datacenter_id: int = 0,
                 *args,
                 **kwargs):
        self._lock = threading.Lock()
        self._next_id = 0

        self.machine_id = machine_id
        self.datacenter_id = datacenter_id

        self.machine_bits = 5
        self.datacenter_bits = 5
        self.max_machine_id = -1 ^ -1 << self.machine_bits
        self.max_datacenter_id = -1 ^ (-1 << self.datacenter_bits)

        self.counter_bits = 12
        self.max_counter_mask = -1 ^ -1 << self.counter_bits

        self.machine_shift = self.counter_bits
        self.datacenter_shift = self.counter_bits + self.machine_bits
        self.timestamp_shift = self.counter_bits + self.machine_bits + self.datacenter_bits

        self.twepoch = int(time.mktime(time.strptime('2019-01-01 00:00:00', "%Y-%m-%d %H:%M:%S")))
        self.last_timestamp = -1
        self.current_timestamp = lambda: int(datetime.now().timestamp() * 1000)

    # def _get_timestamp(self) -> int:
    #     return int(datetime.now().timestamp() * 1000)

    def _get_next_timestamp(self, last_timestamp) -> int:
        timestamp = self.current_timestamp()
        while timestamp <= last_timestamp:
            timestamp = self.current_timestamp()
        return timestamp

    def next(self) -> int:
        with self._lock:
            timestamp = self.current_timestamp()
            if self.last_timestamp == timestamp:
                self._next_id = (self._next_id + 1) & self.max_counter_mask
                if self._next_id == 0:
                    timestamp = self._get_next_timestamp(self.last_timestamp)
            else:
                self._next_id = 0

            if timestamp < self.last_timestamp:
                raise ValueError(
                    'the current timestamp is smaller than the last timestamp')

            self.last_timestamp = timestamp
            uuid = ((timestamp - self.twepoch) << self.timestamp_shift) \
                    | (self.datacenter_id << self.datacenter_shift) \
                    | (self.machine_id << self.machine_shift) \
                    | self._next_id
            return uuid



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import List, Tuple

import numpy as np

from ..base import BaseTextEncoder
from ...helper import batching, as_numpy_array


class FlairEncoder(BaseTextEncoder):
    is_trained = True

    def __init__(self,
                 word_embedding: str = 'glove',
                 flair_embeddings: Tuple[str] = ('news-forward', 'news-backward'),
                 pooling_strategy: str = 'mean', *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.word_embedding = word_embedding
        self.flair_embeddings = flair_embeddings
        self.pooling_strategy = pooling_strategy

    def post_init(self):
        from flair.embeddings import DocumentPoolEmbeddings, WordEmbeddings, FlairEmbeddings
        self._flair = DocumentPoolEmbeddings(
            [WordEmbeddings(self.word_embedding),
             FlairEmbeddings(self.flair_embeddings[0]),
             FlairEmbeddings(self.flair_embeddings[1])],
            pooling=self.pooling_strategy)

    @batching
    @as_numpy_array
    def encode(self, text: List[str], *args, **kwargs) -> np.ndarray:
        from flair.data import Sentence
        import torch
        # tokenize text
        batch_tokens = [Sentence(v) for v in text]
        self._flair.embed(batch_tokens)
        return torch.stack([v.embedding for v in batch_tokens]).detach()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import datetime
import io
import os
import subprocess as sp
from datetime import timedelta
from itertools import product
from typing import List, Callable

import numpy as np

from ..helper import set_logger

logger = set_logger(__name__, True)


def get_video_length(video_path):
    import re
    process = sp.Popen(['ffmpeg', '-i', video_path],
                       stdout=sp.PIPE,
                       stderr=sp.STDOUT)
    stdout, _ = process.communicate()
    stdout = str(stdout)
    matches = re.search(r"Duration:\s{1}(?P<hours>\d+?):(?P<minutes>\d+?):(?P<seconds>\d+\.\d+?),", stdout,
                        re.DOTALL).groupdict()
    h = float(matches['hours'])
    m = float(matches['minutes'])
    s = float(matches['seconds'])

    return 3600 * h + 60 * m + s


def get_video_length_from_raw(buffer_data):
    import re
    ffmpeg_cmd = ['ffmpeg', '-i', '-', '-']
    with sp.Popen(ffmpeg_cmd, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE,
                  bufsize=-1, shell=False) as pipe:
        _, stdout = pipe.communicate(buffer_data)
        stdout = stdout.decode()
        matches = re.search(r"Duration:\s{1}(?P<hours>\d+?):(?P<minutes>\d+?):(?P<seconds>\d+\.\d+?),", stdout,
                            re.DOTALL).groupdict()
        h = str(int(matches['hours']))
        m = str(int(matches['minutes']))
        s = str(round(float(matches['seconds'])))
        duration = datetime.datetime.strptime(h + ':' + m + ':' + s, '%H:%M:%S')
    return duration


def get_audio(buffer_data, sample_rate, interval,
              duration) -> List['np.ndarray']:
    import soundfile as sf

    audio_list = []
    start_time = datetime.datetime.strptime('00:00:00', '%H:%M:%S')
    while True:
        if start_time == duration:
            break

        end_time = start_time + timedelta(seconds=interval)
        if end_time > duration:
            end_time = duration
        ffmpeg_cmd = ['ffmpeg', '-i', '-',
                      '-f', 'wav',
                      '-ar', str(sample_rate),
                      '-ss', str(start_time).split(' ')[1],
                      '-to', str(end_time).split(' ')[1],
                      '-']

        # (-f, wav) output bytes in wav format
        # (-ar) sample rate
        # (-) output to stdout pipeline

        with sp.Popen(
                ffmpeg_cmd, stdin=sp.PIPE, stdout=sp.PIPE, bufsize=-1,
                shell=False) as pipe:
            raw_audio, _ = pipe.communicate(buffer_data)
            tmp_stream = io.BytesIO(raw_audio)
            data, sample_rate = sf.read(tmp_stream)
            # has multiple channels, do average
            if len(data.shape) == 2:
                data = np.mean(data, axis=1)
            if data.shape[0] != 0:
                audio_list.append(data)

        start_time = end_time

    return audio_list


def split_mp4_random(video_path, avg_length, max_clip_second=10):
    import random
    l = get_video_length(video_path)
    s = []
    num_part = max(int(l / avg_length), 2)

    while sum(s) < l:
        s.append(random.randint(3, max_clip_second))
    s[-1] = int(l - sum(s[:-1]))
    start = [sum(s[:i]) for i in range(len(s))]

    ts_group = [[] for _ in range(num_part)]

    for i, (_start, _du) in enumerate(zip(start, s)):
        ts_group[i % num_part].append(' -ss {} -t {} -i {} '.format(_start, _du, video_path))

    prefix = os.path.basename(video_path).replace('.mp4', '')
    for i in range(num_part):
        i_len = len(ts_group[i])
        cmd = 'ffmpeg' + ''.join(
            ts_group[i]) + '-filter_complex "{}concat=n={}:v=1:a=1" -strict -2 {}_{}.mp4 -y'.format(
            ''.join(['[{}]'.format(k) for k in range(i_len)]), i_len, prefix, i)
        os.system(cmd)


def split_video_frames(buffer_data: bytes,
                       splitter: str = '__split__'):
    from PIL import Image

    chunks = buffer_data.split(splitter.encode())
    return [np.array(Image.open(io.BytesIO(chunk))) for chunk in chunks]


def get_gif(images: 'np.ndarray', fps=10):
    cmd = ['ffmpeg', '-y',
           '-f', 'rawvideo',
           '-vcodec', 'rawvideo',
           '-r', '%.02f' % fps,
           '-s', '%dx%d' % (images[0].shape[1], images[0].shape[0]),
           '-pix_fmt', 'rgb24',
           '-i', '-',
           '-filter_complex', '[0:v]split[x][z];[z]palettegen[y];[x]fifo[x];[x][y]paletteuse',
           '-r', '%.02f' % fps,
           '-f', 'gif',
           '-']
    with sp.Popen(cmd, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, bufsize=-1, shell=False) as pipe:
        for image in images:
            pipe.stdin.write(image.tostring())
        out, _ = pipe.communicate()
    return out


def block_descriptor(image: 'np.ndarray',
                     descriptor_fn: Callable,
                     num_blocks: int = 3) -> 'np.ndarray':
    h, w, _ = image.shape  # find shape of image and channel
    block_h = int(np.ceil(h / num_blocks))
    block_w = int(np.ceil(w / num_blocks))

    descriptors = []
    for i in range(0, h, block_h):
        for j in range(0, w, block_w):
            block = image[i:i + block_h, j:j + block_w]
            descriptors.extend(descriptor_fn(block))

    return np.array(descriptors)


def pyramid_descriptor(image: 'np.ndarray',
                       descriptor_fn: Callable,
                       max_level: int = 2) -> 'np.ndarray':
    descriptors = []
    for level in range(max_level + 1):
        num_blocks = 2 ** level
        descriptors.extend(block_descriptor(image, descriptor_fn, num_blocks))

    return np.array(descriptors)


def rgb_histogram(image: 'np.ndarray') -> 'np.ndarray':
    import cv2

    _, _, c = image.shape
    hist = [
        cv2.calcHist([image], [i], None, [256], [0, 256]) for i in range(c)
    ]
    # normalize hist
    hist = np.array([h / np.sum(h) for h in hist]).flatten()
    return hist


def hsv_histogram(image: 'np.ndarray') -> 'np.ndarray':
    import cv2

    _, _, c = image.shape
    hsv = cv2.cvtColor(image, cv2.COLOR_RGB2HSV)

    # sizes = [180, 256, 256]
    # ranges = [(0, 180), (0, 256), (0, 256)]

    # hist = [
    #     cv2.calcHist([hsv], [i], None, [sizes[i]], ranges[i]) for i in range(c)
    # ]

    hist = [cv2.calcHist([hsv], [i], None, [256], [0, 256]) for i in range(c)]
    # normalize hist
    hist = np.array([h / np.sum(h) for h in hist]).flatten()
    return hist


def canny_edge(image: 'np.ndarray', **kwargs) -> 'np.ndarray':
    import cv2

    sigma = kwargs.get('sigma', 0.5)
    gauss_kernel = kwargs.get('gauss_kernel', (9, 9))
    l2_gradient = kwargs.get('l2_gradient', True)

    image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    # apply automatic Canny edge detection using the computed median
    v = np.median(image)
    low_threshold = ((1.0 - sigma) * v).astype("float32")
    high_threshold = ((1.0 + sigma) * v).astype("float32")
    tmp_image = cv2.GaussianBlur(image, gauss_kernel, 1.2)
    edge_image = cv2.Canny(tmp_image, low_threshold, high_threshold, L2gradient=l2_gradient)
    return edge_image


def phash_descriptor(image: 'np.ndarray'):
    from PIL import Image
    import imagehash

    image = Image.fromarray(image)
    return imagehash.phash(image)


def compute_descriptor(image: 'np.ndarray',
                       method: str = 'rgb_histogram',
                       **kwargs) -> 'np.array':
    funcs = {
        'rgb_histogram': rgb_histogram,
        'hsv_histogram': hsv_histogram,
        'canny_edge': lambda image: canny_edge(image, **kwargs),
        'block_rgb_histogram': lambda image: block_descriptor(image, rgb_histogram, kwargs.get('num_blocks', 3)),
        'block_hsv_histogram': lambda image: block_descriptor(image, hsv_histogram, kwargs.get('num_blocks', 3)),
        'pyramid_rgb_histogram': lambda image: pyramid_descriptor(image, rgb_histogram, kwargs.get('max_level', 2)),
        'pyramid_hsv_histogram': lambda image: pyramid_descriptor(image, hsv_histogram, kwargs.get('max_level', 2)),
    }
    return funcs[method](image)


def compare_ecr(descriptors: List['np.ndarray'], **kwargs) -> List[float]:
    import cv2

    """ Apply the Edge Change Ratio Algorithm"""
    dilate_rate = kwargs.get('dilate_rate', 5)
    neigh_avg = kwargs.get('neigh_avg', 2)
    divd = lambda x, y: 0 if y == 0 else x / y

    dicts = []
    inv_dilate = []
    sum_disc = []
    for descriptor in descriptors:
        sum_disc.append(np.sum(descriptor))
        inv_dilate.append(255 - cv2.dilate(descriptor, np.ones((dilate_rate, dilate_rate))))

    for i in range(1, len(descriptors)):
        dict_0 = divd(float(np.sum(descriptors[i - 1] & inv_dilate[i])), float(sum_disc[i - 1]))
        dict_1 = divd(float(np.sum(descriptors[i] & inv_dilate[i - 1])), float(sum_disc[i]))
        tmp_dict = max(dict_0, dict_1)
        if i > 10:
            dict_0 = divd(float(np.sum(descriptors[i - 10] & inv_dilate[i])), float(sum_disc[i - 10]))
            dict_1 = divd(float(np.sum(descriptors[i] & inv_dilate[i - 10])), float(sum_disc[i]))
            tmp_dict *= (1 + max(dict_0, dict_1))
        dicts.append(tmp_dict)

    for _ in range(neigh_avg):
        tmp_dict = []
        for i in range(1, len(dicts) - 1):
            tmp_dict.append(max(dicts[i - 1], dicts[i], dicts[i + 1]))
        dicts = tmp_dict.copy()

    return dicts


def compare_descriptor(descriptor1: 'np.ndarray',
                       descriptor2: 'np.ndarray',
                       metric: str = 'chisqr') -> float:
    import cv2

    dist_metric = {
        'correlation': cv2.HISTCMP_CORREL,
        'chisqr': cv2.HISTCMP_CHISQR,
        'chisqr_alt': cv2.HISTCMP_CHISQR_ALT,
        'intersection': cv2.HISTCMP_INTERSECT,
        'bhattacharya': cv2.HISTCMP_BHATTACHARYYA,
        'hellinguer': cv2.HISTCMP_HELLINGER,
        'kl_div': cv2.HISTCMP_KL_DIV
    }

    return cv2.compareHist(descriptor1, descriptor2, dist_metric[metric])


def kmeans_algo(distances: List[float], **kwargs) -> List[int]:
    from sklearn.cluster import KMeans
    clt = KMeans(n_clusters=2)
    clt.fit(distances)

    num_frames = len(distances) + 1
    # select which cluster includes shot frames
    big_center = np.argmax(clt.cluster_centers_)

    shots = []
    shots.append(0)
    for i in range(0, len(clt.labels_)):
        if big_center == clt.labels_[i]:
            shots.append((i + 1))
    if shots[-1] < num_frames:
        shots.append(num_frames)
    else:
        shots[-1] = num_frames
    return shots


def check_motion(prev_dists: List[float], cur_dist: float, motion_threshold: float = 0.75):
    """ Returns a boolean value to decide if the peak is due to a motion"""
    close_peaks = 0
    # We observe the a defined number of frames before the peak
    for dist in prev_dists:
        if dist > cur_dist * motion_threshold:
            close_peaks += 1
    if close_peaks >= len(prev_dists) / 2:
        return True
    else:
        return False


def thre_algo(distances: List[float], **kwargs) -> List[int]:
    # now threshold algo not support motion
    kwargs['motion_step'] = 0
    return motion_algo(distances, **kwargs)


def motion_algo(distances: List[float], **kwargs) -> List[int]:
    import peakutils

    threshold = kwargs.get('threshold', 0.6)
    min_dist = kwargs.get('min_dist', 10)
    motion_step = kwargs.get('motion_step', 15)
    neigh_avg = kwargs.get('neigh_avg', 2)

    shots = []
    num_frames = len(distances) + 2 * neigh_avg + 1
    p = peakutils.indexes(np.array(distances).astype('float32'), thres=threshold, min_dist=min_dist) if len(distances) else []
    if len(p) == 0:
        return [0, num_frames]

    shots.append(0)
    shots.append(p[0] + neigh_avg + 1)
    for i in range(1, len(p)):
        # We check that the peak is not due to a motion in the image
        valid_dist = not motion_step or not check_motion(distances[p[i]-motion_step:p[i]], distances[p[i]])
        if valid_dist:
            shots.append(p[i] + neigh_avg + 1)
    if shots[-1] < num_frames - min_dist:
        shots.append(num_frames)
    elif shots[-1] > num_frames:
        shots[-1] = num_frames
    return shots


def detect_peak_boundary(distances: List[float],
                         method: str = 'kmeans',
                         **kwargs) -> List[int]:
    detect_method = {
        'kmeans': kmeans_algo,
        'threshold': thre_algo,
        'motion': motion_algo
    }

    if method in detect_method.keys():
        return detect_method[method](distances, **kwargs)
    else:
        logger.error("detect video shot by [%s] not implemented! Please use threshold, kmeans or motion!" % method)


def torch_transform(img):
    import torchvision.transforms as transforms
    return transforms.Compose([transforms.ToTensor(),
                               transforms.Normalize(mean=(0.5, 0.5, 0.5), std=(0.5, 0.5, 0.5))])(img)


def get_all_subarea(img):
    x_list = [0, img.size[0] / 3, 2 * img.size[0] / 3, img.size[0]]
    y_list = [0, img.size[1] / 3, 2 * img.size[1] / 3, img.size[1]]

    index = [[x, y, x + 1, y + 1] for [x, y] in product(range(len(x_list) - 1), range(len(y_list) - 1))]
    all_subareas = [[x_list[idx[0]], y_list[idx[1]], x_list[idx[2]], y_list[idx[3]]] for idx in index]
    return all_subareas, index



import os
import unittest

from gnes.encoder.base import BaseEncoder


class TestTorchTransformersEncoder(unittest.TestCase):

    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.dump_path = os.path.join(dirname, 'model.bin')
        self.text_yaml = os.path.join(dirname, 'yaml', 'torch-transformers-encoder.yml')
        self.tt_encoder = BaseEncoder.load_yaml(self.text_yaml)

        self.test_str = []
        with open(os.path.join(dirname, 'sonnets_small.txt')) as f:
            for line in f:
                line = line.strip()
                if line:
                    self.test_str.append(line)

    def test_encoding(self):
        vec = self.tt_encoder.encode(self.test_str)
        self.assertEqual(vec.shape[0], len(self.test_str))
        self.assertEqual(vec.shape[1], 768)

    def test_dump_load(self):
        self.tt_encoder.dump(self.dump_path)

        tt_encoder2 = BaseEncoder.load(self.dump_path)

        vec = tt_encoder2.encode(self.test_str)
        self.assertEqual(vec.shape[0], len(self.test_str))
        self.assertEqual(vec.shape[1], 768)

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import ctypes
import os
import random
from typing import List, Iterator, Tuple
from typing import Optional

import numpy as np
import zmq
from termcolor import colored

from . import gnes_pb2
from ..helper import batch_iterator, default_logger

__all__ = ['RequestGenerator', 'send_message', 'recv_message',
           'blob2array', 'array2blob', 'gnes_pb2', 'add_route', 'add_version']


class RequestGenerator:
    @staticmethod
    def index(data: Iterator[bytes], batch_size: int = 0, doc_type: int = gnes_pb2.Document.TEXT,
              doc_id_start: int = 0, request_id_start: int = 0,
              random_doc_id: bool = False,
              *args, **kwargs):

        for pi in batch_iterator(data, batch_size):
            req = gnes_pb2.Request()
            req.request_id = request_id_start
            for raw_bytes in pi:
                d = req.index.docs.add()
                d.doc_id = doc_id_start if not random_doc_id else random.randint(0, ctypes.c_uint(-1).value)
                d.raw_bytes = raw_bytes
                d.weight = 1.0
                d.doc_type = doc_type
                doc_id_start += 1
            yield req
            request_id_start += 1

    @staticmethod
    def train(data: Iterator[bytes], batch_size: int = 0, doc_type: int = gnes_pb2.Document.TEXT,
              doc_id_start: int = 0, request_id_start: int = 0,
              random_doc_id: bool = False,
              *args, **kwargs):
        for pi in batch_iterator(data, batch_size):
            req = gnes_pb2.Request()
            req.request_id = request_id_start
            for raw_bytes in pi:
                d = req.train.docs.add()
                d.doc_id = doc_id_start if not random_doc_id else random.randint(0, ctypes.c_uint(-1).value)
                d.raw_bytes = raw_bytes
                d.doc_type = doc_type
                if not random_doc_id:
                    doc_id_start += 1
            yield req
            request_id_start += 1
        req = gnes_pb2.Request()
        req.request_id = request_id_start
        req.train.flush = True
        yield req
        request_id_start += 1

    @staticmethod
    def query(query: bytes, top_k: int, request_id_start: int = 0, doc_type: int = gnes_pb2.Document.TEXT, *args,
              **kwargs):
        if top_k <= 0:
            raise ValueError('"top_k: %d" is not a valid number' % top_k)

        req = gnes_pb2.Request()
        req.request_id = request_id_start
        req.search.query.raw_bytes = query
        req.search.query.doc_type = doc_type
        req.search.top_k = top_k
        yield req


def blob2array(blob: 'gnes_pb2.NdArray') -> np.ndarray:
    """
    Convert a blob proto to an array.
    """
    x = np.frombuffer(blob.data, dtype=blob.dtype).copy()
    return x.reshape(blob.shape)


def array2blob(x: np.ndarray) -> 'gnes_pb2.NdArray':
    """Converts a N-dimensional array to blob proto.
    """
    blob = gnes_pb2.NdArray()
    blob.data = x.tobytes()
    blob.shape.extend(list(x.shape))
    blob.dtype = x.dtype.name
    return blob


def router2str(m: 'gnes_pb2.Message') -> str:
    route_str = [r.service for r in m.envelope.routes]
    return colored('▸', 'green').join(route_str)


def add_route(evlp: 'gnes_pb2.Envelope', name: str, identity: str):
    r = evlp.routes.add()
    r.service = name
    r.start_time.GetCurrentTime()
    r.service_identity = identity


def add_version(evlp: 'gnes_pb2.Envelope'):
    from .. import __version__, __proto_version__
    evlp.gnes_version = __version__
    evlp.proto_version = __proto_version__
    evlp.vcs_version = os.environ.get('GNES_VCS_VERSION', '')


def merge_routes(msg: 'gnes_pb2.Message', prev_msgs: List['gnes_pb2.Message']):
    # take unique routes by service identity
    routes = {(r.service + r.service_identity): r for m in prev_msgs for r in m.envelope.routes}
    msg.envelope.ClearField('routes')
    msg.envelope.routes.extend(sorted(routes.values(), key=lambda x: (x.start_time.seconds, x.start_time.nanos)))


def check_msg_version(msg: 'gnes_pb2.Message'):
    from .. import __version__, __proto_version__
    if hasattr(msg.envelope, 'gnes_version'):
        if not msg.envelope.gnes_version:
            # only happen in unittest
            default_logger.warning('incoming message contains empty "gnes_version", '
                                   'you may ignore it in debug/unittest mode. '
                                   'otherwise please check if frontend service set correct version')
        elif __version__ != msg.envelope.gnes_version:
            raise AttributeError('mismatched GNES version! '
                                 'incoming message has GNES version %s, whereas local GNES version %s' % (
                                     msg.envelope.gnes_version, __version__))

    if hasattr(msg.envelope, 'proto_version'):
        if not msg.envelope.proto_version:
            # only happen in unittest
            default_logger.warning('incoming message contains empty "proto_version", '
                                   'you may ignore it in debug/unittest mode. '
                                   'otherwise please check if frontend service set correct version')
        elif __proto_version__ != msg.envelope.proto_version:
            raise AttributeError('mismatched protobuf version! '
                                 'incoming message has protobuf version %s, whereas local protobuf version %s' % (
                                     msg.envelope.proto_version, __proto_version__))

    if hasattr(msg.envelope, 'vcs_version'):
        if not msg.envelope.vcs_version or not os.environ.get('GNES_VCS_VERSION'):
            default_logger.warning('incoming message contains empty "vcs_version", '
                                   'you may ignore it in debug/unittest mode, '
                                   'or if you run gnes OUTSIDE docker container where GNES_VCS_VERSION is unset'
                                   'otherwise please check if frontend service set correct version')
        elif os.environ.get('GNES_VCS_VERSION') != msg.envelope.vcs_version:
            raise AttributeError('mismatched vcs version! '
                                 'incoming message has vcs_version %s, whereas local environment vcs_version is %s' % (
                                     msg.envelope.vcs_version, os.environ.get('GNES_VCS_VERSION')))

    if not hasattr(msg.envelope, 'proto_version') and not hasattr(msg.envelope, 'gnes_version'):
        raise AttributeError('version_check=True locally, '
                             'but incoming message contains no version info in its envelope. '
                             'the message is probably sent from a very outdated GNES version')


def extract_bytes_from_msg(msg: 'gnes_pb2.Message') -> Tuple:
    doc_bytes = []
    chunk_bytes = []
    doc_byte_type = b''
    chunk_byte_type = b''

    docs = msg.request.train.docs or msg.request.index.docs or [msg.request.search.query]
    # for train request
    for d in docs:
        # oneof raw_data {
        #     string raw_text = 5;
        #       NdArray raw_image = 6;
        #       NdArray raw_video = 7;
        #       bytes raw_bytes = 8; // for other types
        # }
        dtype = d.WhichOneof('raw_data') or ''
        doc_byte_type = dtype.encode()
        if dtype == 'raw_bytes':
            doc_bytes.append(d.raw_bytes)
            d.ClearField('raw_bytes')
        elif dtype == 'raw_image':
            doc_bytes.append(d.raw_image.data)
            d.raw_image.ClearField('data')
        elif dtype == 'raw_video':
            doc_bytes.append(d.raw_video.data)
            d.raw_video.ClearField('data')
        elif dtype == 'raw_text':
            doc_bytes.append(d.raw_text.encode())
            d.ClearField('raw_text')

        for c in d.chunks:
            # oneof content {
            # string text = 2;
            # NdArray blob = 3;
            # bytes raw = 7;
            # }
            chunk_bytes.append(c.embedding.data)
            c.embedding.ClearField('data')

            ctype = c.WhichOneof('content') or ''
            chunk_byte_type = ctype.encode()
            if ctype == 'raw':
                chunk_bytes.append(c.raw)
                c.ClearField('raw')
            elif ctype == 'blob':
                chunk_bytes.append(c.blob.data)
                c.blob.ClearField('data')
            elif ctype == 'text':
                chunk_bytes.append(c.text.encode())
                c.ClearField('text')

    return doc_bytes, doc_byte_type, chunk_bytes, chunk_byte_type


def fill_raw_bytes_to_msg(msg: 'gnes_pb2.Message', msg_data: List[bytes]):
    doc_byte_type = msg_data[2].decode()
    chunk_byte_type = msg_data[3].decode()
    doc_bytes_len = int(msg_data[4])
    chunk_bytes_len = int(msg_data[5])

    doc_bytes = msg_data[6:(6 + doc_bytes_len)]
    chunk_bytes = msg_data[(6 + doc_bytes_len):]

    if len(chunk_bytes) != chunk_bytes_len:
        raise ValueError('"chunk_bytes_len"=%d in message, but the actual length is %d' % (
            chunk_bytes_len, len(chunk_bytes)))

    c_idx = 0
    d_idx = 0
    docs = msg.request.train.docs or msg.request.index.docs or [msg.request.search.query]
    for d in docs:
        if doc_bytes and doc_bytes[d_idx]:
            if doc_byte_type == 'raw_bytes':
                d.raw_bytes = doc_bytes[d_idx]
                d_idx += 1
            elif doc_byte_type == 'raw_image':
                d.raw_image.data = doc_bytes[d_idx]
                d_idx += 1
            elif doc_byte_type == 'raw_video':
                d.raw_video.data = doc_bytes[d_idx]
                d_idx += 1
            elif doc_byte_type == 'raw_text':
                d.raw_text = doc_bytes[d_idx].decode()
                d_idx += 1

        for c in d.chunks:
            if chunk_bytes and chunk_bytes[c_idx]:
                c.embedding.data = chunk_bytes[c_idx]
            c_idx += 1

            if chunk_byte_type == 'raw':
                c.raw = chunk_bytes[c_idx]
                c_idx += 1
            elif chunk_byte_type == 'blob':
                c.blob.data = chunk_bytes[c_idx]
                c_idx += 1
            elif chunk_byte_type == 'text':
                c.text = chunk_bytes[c_idx].decode()
                c_idx += 1


def send_message(sock: 'zmq.Socket', msg: 'gnes_pb2.Message', timeout: int = -1,
                 squeeze_pb: bool = False, **kwargs) -> None:
    try:
        if timeout > 0:
            sock.setsockopt(zmq.SNDTIMEO, timeout)
        else:
            sock.setsockopt(zmq.SNDTIMEO, -1)

        if not squeeze_pb:
            sock.send_multipart([msg.envelope.client_id.encode(), msg.SerializeToString()])
        else:
            doc_bytes, doc_byte_type, chunk_bytes, chunk_byte_type = extract_bytes_from_msg(msg)
            # now raw_bytes are removed from message, hoping for faster de/serialization
            sock.send_multipart(
                [msg.envelope.client_id.encode(),  # 0
                 msg.SerializeToString(),  # 1
                 doc_byte_type, chunk_byte_type,  # 2, 3
                 b'%d' % len(doc_bytes), b'%d' % len(chunk_bytes),  # 4, 5
                 *doc_bytes, *chunk_bytes])  # 6, 7
    except zmq.error.Again:
        raise TimeoutError(
            'cannot send message to sock %s after timeout=%dms, please check the following:'
            'is the server still online? is the network broken? are "port" correct? ' % (
                sock, timeout))
    except Exception as ex:
        raise ex
    finally:
        sock.setsockopt(zmq.SNDTIMEO, -1)


def recv_message(sock: 'zmq.Socket', timeout: int = -1, check_version: bool = False, **kwargs) -> Optional[
    'gnes_pb2.Message']:
    try:
        if timeout > 0:
            sock.setsockopt(zmq.RCVTIMEO, timeout)
        else:
            sock.setsockopt(zmq.RCVTIMEO, -1)

        msg = gnes_pb2.Message()
        msg_data = sock.recv_multipart()
        msg.ParseFromString(msg_data[1])
        if check_version:
            check_msg_version(msg)

        # now we have a barebone msg, we need to fill in data
        if len(msg_data) > 2:
            fill_raw_bytes_to_msg(msg, msg_data)
        return msg

    except zmq.error.Again:
        raise TimeoutError(
            'no response from sock %s after timeout=%dms, please check the following:'
            'is the server still online? is the network broken? are "port" correct? ' % (
                sock, timeout))
    except Exception as ex:
        raise ex
    finally:
        sock.setsockopt(zmq.RCVTIMEO, -1)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

from google.protobuf.json_format import MessageToJson, Parse

from ..base import BaseDocIndexer as BDI
from ...proto import gnes_pb2


class DictIndexer(BDI):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._content = {}

    @BDI.update_counter
    def add(self, keys: List[int], docs: List['gnes_pb2.Document'], *args, **kwargs):
        self._content.update({k: MessageToJson(d) for (k, d) in zip(keys, docs)})

    def query(self, keys: List[int], *args, **kwargs) -> List['gnes_pb2.Document']:
        return [Parse(self._content[k], gnes_pb2.Document()) for k in keys]



import copy
import os
import unittest

from gnes.preprocessor.base import BasePreprocessor
from gnes.preprocessor.video.ffmpeg import FFmpegVideoSegmentor
from gnes.proto import gnes_pb2


class TestPartition(unittest.TestCase):
    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.p3_name = 'pipe-gif'
        self.pipeline_path = os.path.join(self.dirname, 'yaml/%s.yml' % self.p3_name)
        self.ffmpeg_yaml_path = os.path.join(self.dirname, 'yaml/preprocessor-ffmpeg2.yml')
        self.video_path = os.path.join(self.dirname, 'videos')
        self.video_bytes = [open(os.path.join(self.video_path, _), 'rb').read()
                            for _ in os.listdir(self.video_path)]

    def test_gif_pipelinepreproces(self):
        d = gnes_pb2.Document()
        d.raw_bytes = self.video_bytes[0]
        d_ = copy.deepcopy(d)

        p3 = FFmpegVideoSegmentor.load_yaml(self.ffmpeg_yaml_path)
        p3.apply(d)

        p4 = BasePreprocessor.load_yaml(self.pipeline_path)
        p4.apply(d_)

        self.assertEqual(len(d.chunks), len(d_.chunks))



import os
import unittest

from gnes.cli.parser import set_client_cli_parser
from gnes.flow import Flow, Service as gfs, FlowBuildLevelMismatch
from gnes.flow.common import BaseIndexFlow, BaseQueryFlow


class TestGNESFlow(unittest.TestCase):

    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.test_file = os.path.join(self.dirname, 'sonnets_small.txt')
        self.yamldir = os.path.join(self.dirname, 'yaml')
        self.dump_flow_path = os.path.join(self.dirname, 'test-flow.bin')
        self.index_args = set_client_cli_parser().parse_args([
            '--mode', 'index',
            '--txt_file', self.test_file,
            '--batch_size', '4'
        ])
        os.unsetenv('http_proxy')
        os.unsetenv('https_proxy')
        self.test_dir = os.path.join(self.dirname, 'test_flow')
        self.indexer1_bin = os.path.join(self.test_dir, 'my_faiss_indexer.bin')
        self.indexer2_bin = os.path.join(self.test_dir, 'my_fulltext_indexer.bin')
        self.encoder_bin = os.path.join(self.test_dir, 'my_transformer.bin')
        if os.path.exists(self.test_dir):
            self.tearDown()
        os.mkdir(self.test_dir)

        os.environ['TEST_WORKDIR'] = self.test_dir

    def tearDown(self):
        for k in [self.indexer1_bin, self.indexer2_bin, self.encoder_bin, self.dump_flow_path]:
            if os.path.exists(k):
                os.remove(k)
        os.rmdir(self.test_dir)

    def test_flow1(self):
        f = (Flow(check_version=False, route_table=True)
             .add(gfs.Router, yaml_path='BaseRouter'))
        g = f.add(gfs.Router, yaml_path='BaseRouter')

        print('f: %r g: %r' % (f, g))
        g.build()
        print(g.to_mermaid())

        f = f.add(gfs.Router, yaml_path='BaseRouter')
        g = g.add(gfs.Router, yaml_path='BaseRouter')

        print('f: %r g: %r' % (f, g))
        f.build()
        print(f.to_mermaid())
        g.build()

    def test_flow1_ctx_empty(self):
        f = (Flow(check_version=False, route_table=True)
             .add(gfs.Router, yaml_path='BaseRouter'))
        with f(backend='process'):
            pass

    def test_flow1_ctx(self):
        flow = (Flow(check_version=False, route_table=False)
                .add(gfs.Router, yaml_path='BaseRouter'))
        with flow(backend='process', copy_flow=True) as f, open(self.test_file) as fp:
            f.index(txt_file=self.test_file, batch_size=4)
            f.train(txt_file=self.test_file, batch_size=4)

        with flow(backend='process', copy_flow=True) as f:
            # change the flow inside build shall fail
            f = f.add(gfs.Router, yaml_path='BaseRouter')
            self.assertRaises(FlowBuildLevelMismatch, f.index, txt_file=self.test_file, batch_size=4)

        print(flow.build(backend=None).to_mermaid())

    def test_flow2(self):
        f = (Flow(check_version=False, route_table=True)
             .add(gfs.Router, yaml_path='BaseRouter')
             .add(gfs.Router, yaml_path='BaseRouter')
             .add(gfs.Router, yaml_path='BaseRouter')
             .add(gfs.Router, yaml_path='BaseRouter')
             .add(gfs.Router, yaml_path='BaseRouter')
             .add(gfs.Router, yaml_path='BaseRouter')
             .add(gfs.Router, yaml_path='BaseRouter')
             .add(gfs.Router, yaml_path='BaseRouter')
             .build(backend=None))
        print(f._service_edges)
        print(f.to_mermaid())

    def test_flow3(self):
        f = (Flow(check_version=False, route_table=True)
             .add(gfs.Router, name='r0', send_to=gfs.Frontend, yaml_path='BaseRouter')
             .add(gfs.Router, name='r1', recv_from=gfs.Frontend, yaml_path='BaseRouter')
             .build(backend=None))
        print(f._service_edges)
        print(f.to_mermaid())

    def test_flow4(self):
        f = (Flow(check_version=False, route_table=True)
             .add(gfs.Router, name='r0', yaml_path='BaseRouter')
             .add(gfs.Router, name='r1', recv_from=gfs.Frontend, yaml_path='BaseRouter')
             .add(gfs.Router, name='reduce', recv_from=['r0', 'r1'], yaml_path='BaseRouter')
             .build(backend=None))
        print(f._service_edges)
        print(f.to_mermaid())

    def test_flow5(self):
        f = (Flow(check_version=False, route_table=True)
             .add(gfs.Preprocessor, name='prep', yaml_path='SentSplitPreprocessor')
             .add(gfs.Encoder, yaml_path='PyTorchTransformers')
             .add(gfs.Indexer, name='vec_idx', yaml_path='NumpyIndexer')
             .add(gfs.Indexer, name='doc_idx', yaml_path='DictIndexer', recv_from='prep')
             .add(gfs.Router, name='sync_barrier', yaml_path='BaseReduceRouter',
                  num_part=2, recv_from=['vec_idx', 'doc_idx'])
             .build(backend=None))
        print(f._service_edges)
        print(f.to_mermaid())
        # f.to_jpg()

    def test_flow_replica_pot(self):
        f = (Flow(check_version=False, route_table=True)
             .add(gfs.Preprocessor, name='prep', yaml_path='SentSplitPreprocessor', replicas=4)
             .add(gfs.Encoder, yaml_path='PyTorchTransformers', replicas=3)
             .add(gfs.Indexer, name='vec_idx', yaml_path='NumpyIndexer', replicas=2)
             .add(gfs.Indexer, name='doc_idx', yaml_path='DictIndexer', recv_from='prep', replicas=2)
             .add(gfs.Router, name='sync_barrier', yaml_path='BaseReduceRouter',
                  num_part=2, recv_from=['vec_idx', 'doc_idx'])
             .build(backend=None))
        print(f.to_mermaid())
        print(f.to_url(left_right=False))
        print(f.to_url(left_right=True))

    def _test_index_flow(self, backend):
        for k in [self.indexer1_bin, self.indexer2_bin, self.encoder_bin]:
            self.assertFalse(os.path.exists(k))

        flow = (Flow(check_version=False, route_table=False)
                .add(gfs.Preprocessor, name='prep', yaml_path='SentSplitPreprocessor')
                .add(gfs.Encoder, yaml_path=os.path.join(self.dirname, 'yaml/flow-transformer.yml'), replicas=3)
                .add(gfs.Indexer, name='vec_idx', yaml_path=os.path.join(self.dirname, 'yaml/flow-vecindex.yml'))
                .add(gfs.Indexer, name='doc_idx', yaml_path=os.path.join(self.dirname, 'yaml/flow-dictindex.yml'),
                     recv_from='prep')
                .add(gfs.Router, name='sync_barrier', yaml_path='BaseReduceRouter',
                     num_part=2, recv_from=['vec_idx', 'doc_idx']))

        with flow.build(backend=backend) as f:
            f.index(txt_file=self.test_file, batch_size=20)

        for k in [self.indexer1_bin, self.indexer2_bin]:
            self.assertTrue(os.path.exists(k))

    def _test_query_flow(self, backend):
        flow = (Flow(check_version=False, route_table=False)
                .add(gfs.Preprocessor, name='prep', yaml_path='SentSplitPreprocessor')
                .add(gfs.Encoder, yaml_path=os.path.join(self.dirname, 'yaml/flow-transformer.yml'), replicas=3)
                .add(gfs.Indexer, name='vec_idx', yaml_path=os.path.join(self.dirname, 'yaml/flow-vecindex.yml'))
                .add(gfs.Router, name='scorer', yaml_path=os.path.join(self.dirname, 'yaml/flow-score.yml'))
                .add(gfs.Indexer, name='doc_idx', yaml_path=os.path.join(self.dirname, 'yaml/flow-dictindex.yml')))

        with flow.build(backend=backend) as f, open(self.test_file, encoding='utf8') as fp:
            f.query(bytes_gen=[v.encode() for v in fp][:3])

    # @unittest.SkipTest
    def test_index_query_flow(self):
        self._test_index_flow('thread')
        self._test_query_flow('thread')

    def test_indexe_query_flow_proc(self):
        self._test_index_flow('process')
        self._test_query_flow('process')

    def test_query_flow_plot(self):
        flow = (Flow(check_version=False, route_table=False)
                .add(gfs.Preprocessor, name='prep', yaml_path='SentSplitPreprocessor', replicas=2)
                .add(gfs.Encoder, yaml_path=os.path.join(self.dirname, 'yaml/flow-transformer.yml'), replicas=3)
                .add(gfs.Indexer, name='vec_idx', yaml_path=os.path.join(self.dirname, 'yaml/flow-vecindex.yml'),
                     replicas=4)
                .add(gfs.Router, name='scorer', yaml_path=os.path.join(self.dirname, 'yaml/flow-score.yml'))
                .add(gfs.Indexer, name='doc_idx', yaml_path=os.path.join(self.dirname, 'yaml/flow-dictindex.yml')))
        print(flow.build(backend=None).to_url())

    def test_flow_add_set(self):
        f = (Flow(check_version=False, route_table=True)
             .add(gfs.Preprocessor, name='prep', yaml_path='SentSplitPreprocessor', replicas=4)
             .add(gfs.Encoder, yaml_path='PyTorchTransformers', replicas=3)
             .add(gfs.Indexer, name='vec_idx', yaml_path='NumpyIndexer', replicas=2)
             .add(gfs.Indexer, name='doc_idx', yaml_path='DictIndexer', recv_from='prep', replicas=2)
             .add(gfs.Router, name='sync_barrier', yaml_path='BaseReduceRouter',
                  num_part=2, recv_from=['vec_idx', 'doc_idx'])
             .build(backend=None))

        print(f.to_url())
        print(f.set('prep', replicas=1).build(backend=None).to_url())
        # make it as query flow

        f1 = (f
              .remove('sync_barrier')
              .remove('doc_idx')
              .set_last_service('vec_idx')
              .add_router('scorer', yaml_path=os.path.join(self.dirname, 'yaml/flow-score.yml'))
              .add_indexer('doc_idx', yaml_path='DictIndexer', replicas=2)
              .build(backend=None))

        print(f1.to_url())

        # another way to convert f to an index flow

        f2 = (f
              .set_last_service('vec_idx')
              .add_router('scorer', yaml_path=os.path.join(self.dirname, 'yaml/flow-score.yml'))
              .set('doc_idx', recv_from='scorer', yaml_path='DictIndexer', replicas=2, clear_old_attr=True)
              .remove('sync_barrier')
              .set_last_service('doc_idx')
              .build(backend=None))

        print(f2.to_url())

        self.assertEqual(f1, f2)

        self.assertNotEqual(f1, f2.add_router('dummy', yaml_path='BaseRouter'))

        print(f1.to_python_code())
        print(f.to_python_code())

        f1.dump(self.dump_flow_path)
        f3 = Flow.load(self.dump_flow_path)
        self.assertEqual(f1, f3)

        print(f1.to_swarm_yaml())

    def test_common_flow(self):
        print(BaseIndexFlow.build(backend=None).to_url())
        print(BaseQueryFlow.build(backend=None).to_url())
